# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""

import configparser
import datetime
import logging
import pathlib
import traceback
import uuid
from math import log10
from typing import List, Optional

from trestle.oscal import OSCAL_VERSION
from trestle.oscal.common import Metadata
from trestle.oscal.common import Property
from trestle.oscal.component import ComponentDefinition
from trestle.oscal.component import ControlImplementation
from trestle.oscal.component import DefinedComponent
from trestle.oscal.component import ImplementedRequirement
from trestle.oscal.component import SetParameter
from trestle.oscal.component import Statement
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome
from trestle.tasks.csv_helper import CsvHelper

logger = logging.getLogger(__name__)


def derive_control_id(control_mapping: str) -> str:
    """Derive control id."""
    rval = control_mapping.split('_smt')[0]
    return rval


def derive_part_id(control_mapping: str) -> str:
    """Derive part id."""
    if '_smt.' in control_mapping:
        rval = control_mapping
    else:
        rval = None
    return rval


class CsvToOscalComponentDefinition(TaskBase):
    """
    Task to create OSCAL ComponentDefinition json.

    Attributes:
        name: Name of the task.
    """

    name = 'csv-to-oscal-cd'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task csv-to-oscal-cd.

        Args:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)
        self.csv_helper = CsvHelper()
        self._timestamp = datetime.datetime.utcnow().replace(microsecond=0).replace(tzinfo=datetime.timezone.utc
                                                                                    ).isoformat()
        self._verbose = False

    def print_info(self) -> None:
        """Print the help string."""
        self.csv_helper.print_info(self.name, 'component_definition')

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an executed outcome."""
        try:
            return self._execute()
        except Exception:
            logger.error(traceback.format_exc())
            return TaskOutcome('failure')

    def _execute(self) -> TaskOutcome:
        """Execute path core."""
        if not self.csv_helper.configure(self):
            return TaskOutcome('failure')
        # verbosity
        quiet = self._config.get('quiet', False)
        verbose = not quiet
        # config output
        odir = self._config.get('output-dir')
        opth = pathlib.Path(odir)
        self._overwrite = self._config.getboolean('output-overwrite', True)
        # insure output dir exists
        opth.mkdir(exist_ok=True, parents=True)
        # calculate output file name & check writability
        oname = 'component-definition.json'
        ofile = opth / oname
        if not self._overwrite and pathlib.Path(ofile).exists():
            logger.warning(f'output: {ofile} already exists')
            return TaskOutcome('failure')
        # namespace
        self._ns = self._config.get('namespace')
        # user-namespace
        self._ns_user = self._config.get('user-namespace')
        # process rows
        self._process_rows()
        # create OSCAL ComponentDefinition
        metadata = Metadata(
            title=self.csv_helper.get_title(),
            last_modified=self._timestamp,
            oscal_version=OSCAL_VERSION,
            version=self.csv_helper.get_version(),
        )
        component_definition = ComponentDefinition(
            uuid=str(uuid.uuid4()),
            metadata=metadata,
            components=self._get_components(),
        )
        # write OSCAL ComponentDefinition to file
        if verbose:
            logger.info(f'output: {ofile}')
        component_definition.oscal_write(pathlib.Path(ofile))
        # issues
        self._report_issues()
        return TaskOutcome('success')

    def _get_component_key(self, row: List[str]) -> str:
        """Get component key."""
        resource = self.csv_helper.get_value(row, 'Resource')
        component_type = self.csv_helper.get_value(row, 'Component_Type')
        return f'{resource}:{component_type}'

    def _get_component(self, row: List[str]) -> None:
        """Get component."""
        key = self._get_component_key(row)
        rval = self._components.get(key)
        if rval is None:
            resource = self.csv_helper.get_value(row, 'Resource')
            type_ = self.csv_helper.get_value(row, 'Component_Type')
            description = ''
            rval = DefinedComponent(
                uuid=str(uuid.uuid4()),
                type=type_,
                title=resource,
                description=description,
                control_implementations=[],
            )
            self._components[key] = rval
            logger.debug(f'created component: {key}')
        return rval

    def _get_control_implementation_key(self, row: List[str]) -> str:
        """Get control implementation key."""
        resource = self.csv_helper.get_value(row, 'Resource')
        component_type = self.csv_helper.get_value(row, 'Component_Type')
        profile_description = self.csv_helper.get_value(row, 'Profile_Description')
        return f'{resource}:{component_type}:{profile_description}'

    def _get_control_implementation(self, row: List[str]) -> None:
        """Get control implementation."""
        key = self._get_control_implementation_key(row)
        rval = self._control_implementations.get(key)
        if rval is None:
            source = self.csv_helper.get_value(row, 'Profile_Reference_URL')
            description = self.csv_helper.get_value(row, 'Profile_Description')
            rval = ControlImplementation(
                uuid=str(uuid.uuid4()),
                source=source,
                description=description,
                implemented_requirements=[],
            )
            self._control_implementations[key] = rval
            component = self._get_component(row)
            component.control_implementations.append(rval)
            logger.debug(f'created control implementation: {key}')
        return rval

    def _create_rule_prop(
        self,
        row_num: int,
        control_implementation: ControlImplementation,
        row: List[str],
        col: str,
        ns: str,
        remarks: str
    ) -> None:
        """Create rule property."""
        value = self.csv_helper.get_value(row, col)
        logger.debug(f'row_num: {row_num} col: {col} value: {value}')
        if value is None or value == '':
            return
        else:
            class_ = self.csv_helper.get_class(col)
            prop = Property(
                name=col,
                value=value,
                ns=ns,
                class_=class_,
                remarks=remarks,
            )
            control_implementation.props.append(prop)

    def _get_rule_definition_key(self, row: List[str]) -> str:
        """Get rule definition key."""
        resource = self.csv_helper.get_value(row, 'Resource')
        component_type = self.csv_helper.get_value(row, 'Component_Type')
        rule_id = self.csv_helper.get_value(row, 'Rule_Id')
        return f'{resource}:{component_type}:{rule_id}'

    def _register_rule_set(self, row_num: int, key: str, props: List[Property]):
        """Register rule set."""
        self._rule_definitions[key] = props
        logger.debug(f'{row_num} {key} {props}')

    def _add_parameter(self, row_num: int, row: List[str], control_implementation, ns, remarks) -> None:
        # Parameter, if any
        value = self.csv_helper.get_value(row, 'Parameter_Id')
        if value:
            self._create_rule_prop(row_num, control_implementation, row, 'Parameter_Id', ns, remarks)
            self._create_rule_prop(row_num, control_implementation, row, 'Parameter_Description', ns, remarks)
            self._create_rule_prop(row_num, control_implementation, row, 'Parameter_Value_Alternatives', ns, remarks)
            name = self.csv_helper.get_value(row, 'Parameter_Id')
            value = self.csv_helper.get_value(row, 'Parameter_Default_Value')
            if value == '':
                col = 'Parameter_Default_Value'
                text = f'row: {row} missing expected value for col: {col}'
                raise RuntimeError(text)
            values = value.split(',')
            set_parameter = SetParameter(
                param_id=name,
                values=values,
            )
            if control_implementation.set_parameters is None:
                control_implementation.set_parameters = []
            control_implementation.set_parameters.append(set_parameter)

    def _add_rule_definition(self, row_num: int, row: List[str]) -> None:
        """Add rule definition."""
        # Create rule definition (as properties)
        key = self._get_rule_definition_key(row)
        rule_definition = self._rule_definitions.get(key)
        if rule_definition is None:
            control_implementation = self._get_control_implementation(row)
            ns = self._ns
            ns_user = self._ns_user
            fill_sz = int(log10(self.csv_helper.row_count())) + 1
            remarks = f'rule_set_{str(row_num).zfill(fill_sz)}'
            if control_implementation.props is None:
                control_implementation.props = []
            for column_name in self.csv_helper.get_filtered_required_column_names():
                self._create_rule_prop(row_num, control_implementation, row, column_name, ns, remarks)
            for column_name in self.csv_helper.get_filtered_optional_column_names():
                self._create_rule_prop(row_num, control_implementation, row, column_name, ns, remarks)
            # Parameter, if any
            self._add_parameter(row_num, row, control_implementation, ns, remarks)
            # user props
            for col_name in self.csv_helper.get_user_column_names():
                value = self.csv_helper.get_value(row, col_name)
                if not value:
                    continue
                self._create_rule_prop(row_num, control_implementation, row, col_name, ns_user, remarks)
            # Rule set created
            self._register_rule_set(row_num, key, control_implementation.props)
        else:
            text = f'row: {row_num} rule definition: "{key}" already exists?'
            raise RuntimeError(text)

    def _get_implemented_requirement_key(self, control_id: str, row: List[str]) -> str:
        """Get implemented requirement key."""
        resource = self.csv_helper.get_value(row, 'Resource')
        component_type = self.csv_helper.get_value(row, 'Component_Type')
        return f'{resource}:{component_type}:{control_id}'

    def _get_implemented_requirement(self, control_id: str, row: List[str]) -> None:
        """Get implemented requirement."""
        key = self._get_implemented_requirement_key(control_id, row)
        rval = self._implemented_requirements.get(key)
        if rval is None:
            rval = ImplementedRequirement(
                uuid=str(uuid.uuid4()),
                control_id=control_id,
                description='',
            )
            self._implemented_requirements[key] = rval
            control_implementation = self._get_control_implementation(row)
            control_implementation.implemented_requirements.append(rval)
            logger.debug(f'created implemented requirement: {key}')
        return rval

    def _get_statement(self, implemented_requirement: ImplementedRequirement, part_id: str) -> Statement:
        """Get statement."""
        rval = None
        if implemented_requirement.statements is None:
            implemented_requirement.statements = []
        for statement in implemented_requirement.statements:
            if statement.statement_id == part_id:
                rval = statement
                break
        if rval is None:
            statement = Statement(statement_id=part_id, uuid=str(uuid.uuid4()), description='', props=[])
            implemented_requirement.statements.append(statement)
        return statement

    def _add_rule_implementation(self, control_mapping: str, row: List[str]) -> None:
        """Add rule implementation."""
        # Create rule implementation (as property)
        name = 'Rule_Id'
        prop = Property(
            name=name,
            value=self.csv_helper.get_value(row, name),
            ns=self._ns,
            class_=self.csv_helper.get_class(name),
        )
        # Find or create implementation requirement
        control_id = derive_control_id(control_mapping)
        implemented_requirement = self._get_implemented_requirement(control_id, row)
        part_id = derive_part_id(control_mapping)
        # Add rule to implementation requirement as property or statement
        if part_id is None:
            if implemented_requirement.props is None:
                implemented_requirement.props = []
            implemented_requirement.props.append(prop)
        else:
            statement = self._get_statement(implemented_requirement, part_id)
            statement.props.append(prop)

    def _validate_columns(self, row) -> None:
        """Validate columns."""
        for col in self.csv_helper.get_required_column_names():
            value = self.csv_helper.get_value(row, col)
            if value is None or value == '':
                text = f'row: {row} missing expected value for col: {col}'
                raise RuntimeError(text)

    def _process_rows(self) -> None:
        """Process rows."""
        self._rule_definitions = {}
        self._implemented_requirements = {}
        self._control_implementations = {}
        self._components = {}
        for row_num, row in enumerate(self.csv_helper.row_generator()):
            control_mappings = self.csv_helper.get_value(row, 'Control_Mappings').split()
            if len(control_mappings):
                self._validate_columns(row)
                for control_mapping in control_mappings:
                    self._add_rule_implementation(control_mapping, row)
                self._add_rule_definition(row_num, row)

    def _get_components(self) -> List[DefinedComponent]:
        """Get components."""
        rval = []
        for key in self._components.keys():
            rval.append(self._components[key])
        return rval

    def _report_issues(self) -> None:
        """Report issues."""
        self.csv_helper.report_issues()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Template Versioning tests."""
import pathlib

import pytest

from trestle.common.err import TrestleError
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION, TEMPLATE_VERSION_HEADER
from trestle.core.commands.author.versioning.template_versioning import TemplateVersioning
from trestle.core.draw_io import DrawIO
from trestle.core.markdown.markdown_api import MarkdownAPI


def test_template_version_folder_update(tmp_path: pathlib.Path) -> None:
    """Test template folder update from old structure to new."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')
    task_path.mkdir(parents=True)

    old_template = task_path.joinpath('template.md')
    old_template.touch()

    old_folder = task_path.joinpath('images')
    old_folder.mkdir(parents=True)

    with pytest.raises(TrestleError):
        TemplateVersioning.update_template_folder_structure(old_template)

    # make sure file and folder gets put to first version
    TemplateVersioning.update_template_folder_structure(task_path)

    assert task_path.joinpath('0.0.1/').joinpath('template.md').exists()
    assert task_path.joinpath('0.0.1/').joinpath('images').exists()
    assert not old_template.exists()
    assert not old_folder.exists()

    # make sure new file also gets put to the first version
    another_template = task_path.joinpath('missplaced_template.md')
    another_template.touch()

    TemplateVersioning.update_template_folder_structure(task_path)
    assert task_path.joinpath('0.0.1/').joinpath('template.md').exists()
    assert task_path.joinpath('0.0.1/').joinpath('missplaced_template.md').exists()
    assert not another_template.exists()

    # make sure other versions get no modifications
    v2_dir = task_path.joinpath('11.12.113')
    v2_dir.mkdir()

    yet_another_template = task_path.joinpath('new_template.md')
    yet_another_template.touch()
    TemplateVersioning.update_template_folder_structure(task_path)

    assert len(list(v2_dir.iterdir())) == 0
    assert task_path.joinpath('0.0.1/').joinpath('template.md').exists()
    assert task_path.joinpath('0.0.1/').joinpath('missplaced_template.md').exists()
    assert task_path.joinpath('0.0.1/').joinpath('new_template.md').exists()
    assert not yet_another_template.exists()


def test_get_latest_version(tmp_path: pathlib.Path) -> None:
    """Test get latest version of template."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')

    with pytest.raises(TrestleError):
        TemplateVersioning.get_latest_version_for_task(task_path)

    task_path.mkdir(parents=True)

    latest_path, version = TemplateVersioning.get_latest_version_for_task(task_path)
    assert latest_path == task_path.joinpath('0.0.1')
    assert version == '0.0.1'

    v1_dir = task_path.joinpath('0.0.1/')
    v2_dir = task_path.joinpath('11.10.1234/')
    v1_dir.mkdir(parents=True)
    v2_dir.mkdir(parents=True)

    latest_path, version = TemplateVersioning.get_latest_version_for_task(task_path)
    assert latest_path == v2_dir
    assert version == '11.10.1234'

    template_v1 = v1_dir.joinpath('template.md')
    template_v2 = v2_dir.joinpath('template.md')
    template_v1.touch()
    template_v2.touch()

    latest_path, version = TemplateVersioning.get_latest_version_for_task(task_path)
    assert latest_path == v2_dir
    assert version == '11.10.1234'

    with pytest.raises(TrestleError):
        TemplateVersioning.get_latest_version_for_task(template_v1)


def test_get_versioned_template(tmp_path: pathlib.Path) -> None:
    """Test get template of the specified version."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')

    with pytest.raises(TrestleError):
        TemplateVersioning.get_versioned_template_dir(task_path)

    task_path.mkdir(parents=True)

    with pytest.raises(TrestleError):
        TemplateVersioning.get_versioned_template_dir(task_path)

    v1_dir = task_path.joinpath(START_TEMPLATE_VERSION)
    v2_dir = task_path.joinpath('0.1.5/')
    v3_dir = task_path.joinpath('10.02.1/')
    v1_dir.mkdir(parents=True)
    v2_dir.mkdir(parents=True)
    v3_dir.mkdir(parents=True)

    first_version = TemplateVersioning.get_versioned_template_dir(task_path, START_TEMPLATE_VERSION)
    assert first_version == v1_dir

    latest_version = TemplateVersioning.get_versioned_template_dir(task_path)
    assert latest_version == v3_dir

    second_version = TemplateVersioning.get_versioned_template_dir(task_path, '0.1.5')
    assert second_version == v2_dir

    with pytest.raises(TrestleError):
        TemplateVersioning.get_versioned_template_dir(task_path, '6.7.8')


def test_write_versioned_template(tmp_path: pathlib.Path) -> None:
    """Test writing a template to the folder."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')

    with pytest.raises(TrestleError):
        TemplateVersioning.get_versioned_template_dir(task_path)

    task_path.mkdir(parents=True)
    tmp_path_01 = task_path.joinpath('0.0.1')
    tmp_path_02 = task_path.joinpath('0.0.2')
    tmp_path_01.mkdir(parents=True)
    tmp_path_02.mkdir(parents=True)
    template = tmp_path_01.joinpath('template.md')

    TemplateVersioning.write_versioned_template('template.md', tmp_path_01, template, None)

    assert task_path.joinpath(START_TEMPLATE_VERSION).joinpath('template.md').exists()
    assert template.exists()

    template2 = tmp_path_01.joinpath('template2.md')
    template3 = tmp_path_02.joinpath('template3.md')
    template4 = tmp_path_02.joinpath('template4.md')
    TemplateVersioning.write_versioned_template('template.md', tmp_path_01, template2, '0.0.1')
    TemplateVersioning.write_versioned_template('template.md', tmp_path_02, template3, '0.0.2')
    TemplateVersioning.write_versioned_template('template.md', tmp_path_02, template4, None)

    assert tmp_path_01.joinpath('template.md').exists()
    assert tmp_path_01.joinpath('template2.md').exists()
    assert tmp_path_02.joinpath('template3.md').exists()
    assert tmp_path_02.joinpath('template4.md').exists()
    assert template.exists()

    md_api = MarkdownAPI()
    header, _ = md_api.processor.read_markdown_wo_processing(tmp_path_02.joinpath('template3.md'))

    assert header[TEMPLATE_VERSION_HEADER] == '0.0.2'

    template_drawio = tmp_path_02.joinpath('template.drawio')
    TemplateVersioning.write_versioned_template('template.drawio', tmp_path_02, template_drawio, '0.0.2')

    drawio = DrawIO(template_drawio)
    metadata = drawio.get_metadata()[0]

    assert metadata[TEMPLATE_VERSION_HEADER] == '0.0.2'


def test_template_versioning(tmp_path: pathlib.Path) -> None:
    """Full on testing of updating the template folder structure and retrieving it back."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')

    task_path.mkdir(parents=True)
    task_path.joinpath('0.0.1').mkdir(parents=True)
    task_path.joinpath('0.0.2').mkdir(parents=True)
    template = task_path.joinpath('0.0.2/template.md')
    template2 = task_path.joinpath('0.0.2/template2.md')

    old_template = task_path.joinpath('old_template.md')
    old_template.touch()

    TemplateVersioning.update_template_folder_structure(task_path)
    TemplateVersioning.write_versioned_template('template.md', task_path.joinpath('0.0.2'), template, '0.0.2')
    TemplateVersioning.write_versioned_template('template.md', task_path.joinpath('0.0.2'), template2, None)

    v1_dir = TemplateVersioning.get_versioned_template_dir(task_path, START_TEMPLATE_VERSION)
    latest_dir, version = TemplateVersioning.get_latest_version_for_task(task_path)
    v2_dir = TemplateVersioning.get_versioned_template_dir(task_path, '0.0.2')

    assert not old_template.exists()
    assert task_path.joinpath(START_TEMPLATE_VERSION).joinpath('old_template.md').exists()
    assert task_path.joinpath('0.0.2').joinpath('template.md').exists()
    assert task_path.joinpath('0.0.2').joinpath('template2.md').exists()
    assert v1_dir.exists()
    assert v2_dir == latest_dir
    assert version == '0.0.2'


def test_valid_version() -> None:
    """Test is valid verion."""
    assert TemplateVersioning.is_valid_version('0.0.1')
    assert not TemplateVersioning.is_valid_version("'0.0.1'")
    assert not TemplateVersioning.is_valid_version('not_valid.0.0.1')
    assert not TemplateVersioning.is_valid_version('0.1')
    assert not TemplateVersioning.is_valid_version('1')
    assert not TemplateVersioning.is_valid_version('0.0.0.1')


def test_empty_folder_is_not_created(tmp_path: pathlib.Path) -> None:
    """Test that empty folder is not created."""
    task_path = tmp_path.joinpath('trestle/author/sample_task/')
    task_path.mkdir(parents=True)

    # run template folder update with empty task
    TemplateVersioning.update_template_folder_structure(task_path)

    assert not task_path.joinpath('0.0.1/').exists()

    # now create other version of the template
    newest_version_path = task_path.joinpath('1.0.0')
    newest_version_path.mkdir(parents=True)

    old_template = newest_version_path.joinpath('template.md')
    old_template.touch()

    old_folder = newest_version_path.joinpath('images')
    old_folder.mkdir(parents=True)

    # run template folder update with newest version
    TemplateVersioning.update_template_folder_structure(task_path)

    # assert everything is stayed the same
    assert not task_path.joinpath('0.0.1/').exists()
    assert old_folder.exists()
    assert old_template.exists()



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create ssp from catalog and profile."""

import argparse
import logging
import pathlib
from typing import Dict, List, Optional, Set

from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

import trestle.core.generators as gens
import trestle.core.generic_oscal as generic
import trestle.oscal.common as com
import trestle.oscal.profile as prof
import trestle.oscal.ssp as ossp
from trestle.common import const, file_utils, log
from trestle.common.err import TrestleError, handle_generic_command_exception
from trestle.common.list_utils import as_list, none_if_empty
from trestle.common.load_validate import load_validate_model_name
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog_interface import CatalogInterface
from trestle.core.commands.author.common import AuthorCommonCommand
from trestle.core.commands.author.profile import sections_to_dict
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_reader import ControlReader
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver

logger = logging.getLogger(__name__)


class SSPGenerate(AuthorCommonCommand):
    """Generate SSP in markdown form from a Profile."""

    name = 'ssp-generate'

    def _init_arguments(self) -> None:
        file_help_str = 'Name of the profile model in the trestle workspace'
        self.add_argument('-p', '--profile', help=file_help_str, required=True, type=str)
        self.add_argument('-o', '--output', help=const.HELP_MARKDOWN_NAME, required=True, type=str)
        self.add_argument('-y', '--yaml-header', help=const.HELP_YAML_PATH, required=False, type=str)
        self.add_argument(
            '-ohv',
            '--overwrite-header-values',
            help=const.HELP_OVERWRITE_HEADER_VALUES,
            required=False,
            action='store_true',
            default=False
        )
        sections_help_str = (
            'Comma separated list of section:alias pairs.  Provides mapping of short names to long for markdown.'
        )
        self.add_argument('-s', '--sections', help=sections_help_str, required=False, type=str)
        allowed_sections_help_str = (
            'Comma separated list of section short names to include in the markdown.  Others will not appear.'
        )
        self.add_argument('-as', '--allowed-sections', help=allowed_sections_help_str, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = args.trestle_root
            if not file_utils.is_directory_name_allowed(args.output):
                raise TrestleError(f'{args.output} is not an allowed directory name')

            profile_path = trestle_root / f'profiles/{args.profile}/profile.json'

            yaml_header: dict = {}
            if args.yaml_header:
                try:
                    logging.debug(f'Loading yaml header file {args.yaml_header}')
                    yaml = YAML()
                    yaml_header = yaml.load(pathlib.Path(args.yaml_header).open('r'))
                except YAMLError as e:
                    raise TrestleError(f'YAML error loading yaml header for ssp generation: {e}')

            markdown_path = trestle_root / args.output

            profile_resolver = ProfileResolver()

            # in ssp context we want to see missing value warnings
            resolved_catalog = profile_resolver.get_resolved_profile_catalog(
                trestle_root, profile_path, show_value_warnings=True
            )
            catalog_interface = CatalogInterface(resolved_catalog)

            sections_dict: Dict[str, str] = {}
            if args.sections:
                sections_dict = sections_to_dict(args.sections)
                if const.STATEMENT in sections_dict:
                    raise TrestleError('Statement is not allowed as a section name.')
                # add any existing sections from the controls but only have short names
                control_section_short_names = catalog_interface.get_sections()
                for short_name in control_section_short_names:
                    if short_name not in sections_dict:
                        sections_dict[short_name] = short_name
                logger.debug(f'ssp sections dict: {sections_dict}')

            context = ControlContext.generate(ContextPurpose.SSP, True, trestle_root, markdown_path)
            context.yaml_header = yaml_header
            context.sections_dict = sections_dict
            context.prompt_responses = True
            context.overwrite_header_values = args.overwrite_header_values
            context.allowed_sections = args.allowed_sections

            catalog_interface.write_catalog_as_markdown(context, catalog_interface.get_statement_part_id_map(False))

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while writing markdown from catalog')


class SSPAssemble(AuthorCommonCommand):
    """Assemble markdown files of controls into an SSP json file."""

    name = 'ssp-assemble'

    def _init_arguments(self) -> None:
        name_help_str = (
            'Optional name of the ssp model in the trestle workspace that is being modified.  '
            'If not provided the output name is used.'
        )
        self.add_argument('-n', '--name', help=name_help_str, required=False, type=str)
        file_help_str = 'Name of the input markdown file directory'
        self.add_argument('-m', '--markdown', help=file_help_str, required=True, type=str)
        output_help_str = 'Name of the output generated json SSP'
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)

    def _merge_imp_reqs(self, ssp: ossp.SystemSecurityPlan, imp_reqs: List[ossp.ImplementedRequirement]) -> None:
        """
        Merge the new imp_reqs into the ssp and optionally regenerate uuids.

        If a statement has same id and same by_comp uuid as ssp, use the ssp version with new description.
        Otherwise just insert the statement.
        When the statement was loaded it had access to the current components so the uuids should match.
        """
        id_map: Dict[str, Dict[str, ossp.Statement]] = {}
        control_map: Dict[str, ossp.ImplementedRequirement] = {}
        for imp_req in ssp.control_implementation.implemented_requirements:
            control_map[imp_req.control_id] = imp_req
            for statement in imp_req.statements:
                for by_comp in statement.by_components:
                    id_ = statement.statement_id
                    if id_ not in id_map:
                        id_map[id_] = {}
                    id_map[id_][by_comp.component_uuid] = statement

        for imp_req in imp_reqs:
            if imp_req.control_id in control_map:
                imp_req.uuid = control_map[imp_req.control_id].uuid
            for statement in as_list(imp_req.statements):
                id_ = statement.statement_id
                # for each statement id match the statement per component to the original
                if id_ in id_map:
                    comp_dict = id_map[id_]
                    for by_comp in as_list(statement.by_components):
                        if by_comp.component_uuid in comp_dict:
                            statement.uuid = comp_dict[by_comp.component_uuid].uuid
                            for orig_by_comp in as_list(comp_dict[by_comp.component_uuid].by_components):
                                if orig_by_comp.component_uuid == by_comp.component_uuid:
                                    by_comp.uuid = orig_by_comp.uuid
                                    break

        changed = ssp.control_implementation.implemented_requirements != imp_reqs
        ssp.control_implementation.implemented_requirements = imp_reqs
        return changed

    def _generate_roles_in_metadata(self, ssp: ossp.SystemSecurityPlan) -> bool:
        """Find all roles referenced by imp reqs and create role in metadata as needed."""
        metadata = ssp.metadata
        metadata.roles = as_list(metadata.roles)
        known_role_ids = [role.id for role in metadata.roles]
        changed = False
        for imp_req in ssp.control_implementation.implemented_requirements:
            role_ids = [resp_role.role_id for resp_role in as_list(imp_req.responsible_roles)]
            for role_id in role_ids:
                if role_id not in known_role_ids:
                    role = com.Role(id=role_id, title=role_id)
                    metadata.roles.append(role)
                    known_role_ids.append(role_id)
                    changed = True
        metadata.roles = none_if_empty(metadata.roles)
        return changed

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = pathlib.Path(args.trestle_root)

            md_path = trestle_root / args.markdown

            # the original, reference ssp name defaults to same as output if name not specified
            # thus in cyclic editing you are reading and writing same json ssp
            orig_ssp_name = args.output
            if args.name:
                orig_ssp_name = args.name
            new_ssp_name = args.output

            new_file_content_type = FileContentType.JSON

            # if output ssp already exists, load it to see if new one is different
            existing_ssp: Optional[ossp.SystemSecurityPlan] = None
            new_ssp_path = ModelUtils.full_path_for_top_level_model(trestle_root, new_ssp_name, ossp.SystemSecurityPlan)
            if new_ssp_path:
                _, _, existing_ssp = ModelUtils.load_distributed(new_ssp_path, trestle_root)
                new_file_content_type = FileContentType.path_to_content_type(new_ssp_path)

            ssp: ossp.SystemSecurityPlan
            comp_dict: Dict[str, generic.GenericComponent] = {}

            # if orig ssp exists - need to load it rather than instantiate new one
            orig_ssp_path = ModelUtils.full_path_for_top_level_model(
                trestle_root, orig_ssp_name, ossp.SystemSecurityPlan
            )

            context = ControlContext.generate(ContextPurpose.SSP, True, trestle_root, md_path)

            # need to load imp_reqs from markdown but need component first
            if orig_ssp_path:
                # load the existing json ssp
                _, _, ssp = ModelUtils.load_distributed(orig_ssp_path, trestle_root)
                for component in ssp.system_implementation.components:
                    comp_dict[component.title] = generic.GenericComponent.from_system_component(component)
                # read the new imp reqs from markdown and have them reference existing components
                imp_reqs = CatalogInterface.read_catalog_imp_reqs(md_path, comp_dict, context)
                new_imp_reqs = []
                for imp_req in imp_reqs:
                    new_imp_reqs.append(imp_req.as_ssp())
                self._merge_imp_reqs(ssp, new_imp_reqs)
                new_file_content_type = FileContentType.path_to_content_type(orig_ssp_path)
            else:
                # create a sample ssp to hold all the parts
                ssp = gens.generate_sample_model(ossp.SystemSecurityPlan)
                # load the imp_reqs from markdown and create components as needed, referenced by ### headers
                imp_reqs = CatalogInterface.read_catalog_imp_reqs(md_path, comp_dict, context)
                new_imp_reqs = []
                for imp_req in imp_reqs:
                    new_imp_reqs.append(imp_req.as_ssp())

                # create system implementation
                system_imp: ossp.SystemImplementation = gens.generate_sample_model(ossp.SystemImplementation)
                ssp.system_implementation = system_imp

                # create a control implementation to hold the implementated requirements
                control_imp: ossp.ControlImplementation = gens.generate_sample_model(ossp.ControlImplementation)
                control_imp.implemented_requirements = new_imp_reqs
                control_imp.description = const.SSP_SYSTEM_CONTROL_IMPLEMENTATION_TEXT

                # insert the parts into the ssp
                ssp.control_implementation = control_imp
                ssp.system_implementation = system_imp

                # we don't have access to the original profile so we don't know the href
                import_profile: ossp.ImportProfile = gens.generate_sample_model(ossp.ImportProfile)
                import_profile.href = 'REPLACE_ME'
                ssp.import_profile = import_profile

            # now that we know the complete list of needed components, add them to the sys_imp
            # TODO if the ssp already existed then components may need to be removed if not ref'd by imp_reqs
            component_list: List[ossp.SystemComponent] = []
            for comp in comp_dict.values():
                # need to skip the component corresponding to statement level prose
                if comp.title:
                    component_list.append(comp.as_system_component())
            if ssp.system_implementation.components:
                # reconstruct list with same order as existing, but add/remove components as needed
                new_list: List[ossp.SystemComponent] = []
                for comp in ssp.system_implementation.components:
                    if comp in component_list:
                        new_list.append(comp)
                for comp in component_list:
                    if comp not in new_list:
                        new_list.append(comp)
                ssp.system_implementation.components = new_list
            elif component_list:
                ssp.system_implementation.components = component_list
            self._generate_roles_in_metadata(ssp)

            if args.version:
                ssp.metadata.version = com.Version(__root__=args.version)

            if ModelUtils.models_are_equivalent(existing_ssp, ssp):
                logger.info('No changes to assembled ssp so ssp not written out.')
                return CmdReturnCodes.SUCCESS.value

            if args.regenerate:
                ssp, _, _ = ModelUtils.regenerate_uuids(ssp)
            ModelUtils.update_last_modified(ssp)

            # write out the ssp as json
            ModelUtils.save_top_level_model(ssp, trestle_root, new_ssp_name, new_file_content_type)

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while assembling SSP')


class SSPFilter(AuthorCommonCommand):
    """Filter the controls in an ssp based on files included by profile and/or list of component names."""

    name = 'ssp-filter'

    def _init_arguments(self) -> None:
        file_help_str = 'Name of the input ssp'
        self.add_argument('-n', '--name', help=file_help_str, required=True, type=str)
        file_help_str = 'Name of the optional input profile that defines set of controls in filtered ssp'
        self.add_argument('-p', '--profile', help=file_help_str, required=False, type=str)
        output_help_str = 'Name of the output generated SSP'
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)
        comp_help_str = 'Colon-delimited list of component names to include in filtered ssp.'
        self.add_argument('-c', '--components', help=comp_help_str, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = pathlib.Path(args.trestle_root)
            comp_names: Optional[List[str]] = None
            if args.components:
                comp_names = args.components.split(':')
            elif not args.profile:
                logger.warning('You must specify either a profile or list of component names for ssp-filter.')
                return 1

            return self.filter_ssp(
                trestle_root, args.name, args.profile, args.output, args.regenerate, args.version, comp_names
            )
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error generating the filtered ssp')

    def filter_ssp(
        self,
        trestle_root: pathlib.Path,
        ssp_name: str,
        profile_name: str,
        out_name: str,
        regenerate: bool,
        version: Optional[str],
        components: Optional[List[str]] = None
    ) -> int:
        """
        Filter the ssp based on controls included by the profile and/or components and output new ssp.

        Args:
            trestle_root: root directory of the trestle project
            ssp_name: name of the ssp model
            profile_name: name of the optional profile model used for filtering
            out_name: name of the output ssp model with filtered controls
            regenerate: whether to regenerate the uuid's in the ssp
            version: new version for the model
            components: optional list of component names used for filtering

        Returns:
            0 on success, 1 otherwise
        """
        # load the ssp
        ssp: ossp.SystemSecurityPlan
        ssp, _ = load_validate_model_name(trestle_root, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)
        profile_path = ModelUtils.path_for_top_level_model(
            trestle_root, profile_name, prof.Profile, FileContentType.JSON
        )

        if components:
            raw_comp_names = [ControlReader.simplify_name(name) for name in components]
            comp_uuids: List[str] = []
            for component in ssp.system_implementation.components:
                if ControlReader.simplify_name(component.title) in raw_comp_names:
                    comp_uuids.append(component.uuid)
            # imp_reqs can be by comp
            # and imp_reqs can have statements that are by comp
            if comp_uuids:
                new_imp_reqs: List[ossp.ImplementedRequirement] = []
                # these are all required to be present
                for imp_req in ssp.control_implementation.implemented_requirements:
                    new_by_comps: List[ossp.ByComponent] = []
                    # by_comps is optional
                    for by_comp in as_list(imp_req.by_components):
                        if by_comp.component.uuid in comp_uuids:
                            new_by_comps.append(by_comp)
                    imp_req.by_components = none_if_empty(new_by_comps)
                    new_imp_reqs.append(imp_req)
                    new_statements: List[ossp.Statement] = []
                    for statement in as_list(imp_req.statements):
                        new_by_comps: List[ossp.ByComponent] = []
                        for by_comp in as_list(statement.by_components):
                            if by_comp.component_uuid in comp_uuids:
                                new_by_comps.append(by_comp)
                        statement.by_components = none_if_empty(new_by_comps)
                        new_statements.append(statement)
                    imp_req.statements = none_if_empty(new_statements)
                ssp.control_implementation.implemented_requirements = new_imp_reqs
                # now remove any unused components from the ssp
                new_comp_list: List[ossp.SystemComponent] = []
                for comp in ssp.system_implementation.components:
                    if comp.uuid in comp_uuids:
                        new_comp_list.append(comp)
                ssp.system_implementation.components = new_comp_list

        # filter by controls in profile
        if profile_name:
            prof_resolver = ProfileResolver()
            catalog = prof_resolver.get_resolved_profile_catalog(trestle_root, profile_path, show_value_warnings=True)
            catalog_interface = CatalogInterface(catalog)

            # The input ssp should reference a superset of the controls referenced by the profile
            # Need to cull references in the ssp to controls not in the profile
            # Also make sure the output ssp contains imp reqs for all controls in the profile
            control_imp = ssp.control_implementation
            ssp_control_ids: Set[str] = set()

            new_set_params: List[ossp.SetParameter] = []
            for set_param in as_list(control_imp.set_parameters):
                control = catalog_interface.get_control_by_param_id(set_param.param_id)
                if control is not None:
                    new_set_params.append(set_param)
                    ssp_control_ids.add(control.id)
            control_imp.set_parameters = none_if_empty(new_set_params)

            new_imp_requirements: List[ossp.ImplementedRequirement] = []
            for imp_requirement in as_list(control_imp.implemented_requirements):
                control = catalog_interface.get_control(imp_requirement.control_id)
                if control is not None:
                    new_imp_requirements.append(imp_requirement)
                    ssp_control_ids.add(control.id)
            control_imp.implemented_requirements = new_imp_requirements

            # make sure all controls in the profile have implemented reqs in the final ssp
            if not ssp_control_ids.issuperset(catalog_interface.get_control_ids()):
                raise TrestleError('Unable to filter the ssp because the profile references controls not in it.')

            ssp.control_implementation = control_imp

        if version:
            ssp.metadata.version = com.Version(__root__=version)

        existing_ssp_path = ModelUtils.full_path_for_top_level_model(trestle_root, out_name, ossp.SystemSecurityPlan)
        if existing_ssp_path is not None:
            existing_ssp, _ = load_validate_model_name(trestle_root, out_name, ossp.SystemSecurityPlan)
            if ModelUtils.models_are_equivalent(existing_ssp, ssp):
                logger.info('No changes to filtered ssp so ssp not written out.')
                return CmdReturnCodes.SUCCESS.value

        if regenerate:
            ssp, _, _ = ModelUtils.regenerate_uuids(ssp)

        ModelUtils.update_last_modified(ssp)

        ModelUtils.save_top_level_model(ssp, trestle_root, out_name, FileContentType.JSON)

        return CmdReturnCodes.SUCCESS.value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle assemble command."""
import argparse
import os
import pathlib
import shutil
import sys

from _pytest.monkeypatch import MonkeyPatch

import pytest

import trestle.common.err as err
from trestle.cli import Trestle
from trestle.common import const
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.assemble import AssembleCmd
from trestle.core.models.plans import Plan
from trestle.oscal.catalog import Catalog

subcommand_list = const.MODEL_TYPE_LIST


def test_run_and_missing_model(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test _run and test it fails without top level model file."""
    testargs_root = ['trestle', 'assemble']
    for subcommand in subcommand_list:
        test_args = testargs_root + [subcommand] + ['-n', f'my_{subcommand}'] + ['-x', 'json']
        monkeypatch.setattr(sys, 'argv', test_args)
        rc = Trestle().run()
        assert rc != 0


def test_assemble_catalog(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test assembling a catalog."""
    test_data_source = testdata_dir / 'split_merge/step4_split_groups_array/catalogs'
    catalogs_dir = tmp_trestle_dir / 'catalogs'
    mycatalog_dir = catalogs_dir / 'mycatalog'
    # Copy files from test/data/split_merge/step4
    shutil.rmtree(catalogs_dir)
    shutil.rmtree(pathlib.Path('dist'))
    shutil.copytree(test_data_source, catalogs_dir)

    testargs = ['trestle', 'assemble', 'catalog', '-n', 'mycatalog', '-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 0

    # Read assembled model
    actual_model = Catalog.oscal_read(pathlib.Path('dist/catalogs/mycatalog.json'))
    _, _, expected_model = ModelUtils.load_distributed(mycatalog_dir / 'catalog.json', tmp_trestle_dir)

    assert actual_model == expected_model


def test_assemble_not_trestle_project(tmp_empty_cwd: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test failure if not trestle project."""
    testargs = ['trestle', 'assemble', 'catalog', '-n', 'mycatalog', '-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 5


def test_assemble_not_trestle_root(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test execution of assemble from a folder that is not trestle root."""
    os.chdir(pathlib.Path.cwd() / 'catalogs')
    testargs = ['trestle', 'assemble', 'catalog', '-n', 'mycatalog', '-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 1


def test_assemble_execution_failure(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test execution of assemble plan fails."""

    def execute_mock(*args, **kwargs):
        raise err.TrestleError('execution failed')

    test_data_source = testdata_dir / 'split_merge/step4_split_groups_array/catalogs'
    catalogs_dir = pathlib.Path('catalogs/')
    # Copy files from test/data/split_merge/step4
    shutil.rmtree(catalogs_dir)
    shutil.rmtree(pathlib.Path('dist'))
    shutil.copytree(test_data_source, catalogs_dir)
    monkeypatch.setattr(Plan, 'execute', execute_mock)
    with pytest.raises(err.TrestleError):
        AssembleCmd().assemble_model(
            'catalog', argparse.Namespace(trestle_root=tmp_trestle_dir, name='mycatalog', extension='json', verbose=1)
        )


def test_assemble_missing_top_model(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test assembling a catalog."""
    test_data_source = testdata_dir / 'split_merge/step4_split_groups_array/catalogs'
    catalogs_dir = pathlib.Path('catalogs/')
    mycatalog_dir = catalogs_dir / 'mycatalog'
    # Copy files from test/data/split_merge/step4
    shutil.rmtree(catalogs_dir)
    shutil.rmtree(pathlib.Path('dist'))
    shutil.copytree(test_data_source, catalogs_dir)
    (mycatalog_dir / 'catalog.json').unlink()

    testargs = ['trestle', 'assemble', 'catalog', '-n', 'mycatalog', '-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 1


def test_assemble_catalog_all(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test assembling all catalogs in trestle dir."""
    shutil.rmtree(pathlib.Path('dist'))
    catalogs_dir = tmp_trestle_dir / 'catalogs'
    my_names = ['mycatalog1', 'mycatalog2', 'mycatalog3']
    for my_name in my_names:
        test_data_source = testdata_dir / 'split_merge/step4_split_groups_array/catalogs/mycatalog'
        shutil.copytree(test_data_source, catalogs_dir / my_name)

    testargs = ['trestle', 'assemble', 'catalog', '-t', '-x', 'json']
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 0

    # Read assembled model
    for my_name in my_names:
        _, _, expected_model = ModelUtils.load_distributed(catalogs_dir / f'{my_name}/catalog.json', tmp_trestle_dir)
        actual_model = Catalog.oscal_read(pathlib.Path(f'dist/catalogs/{my_name}.json'))
        assert actual_model == expected_model

    testargs = ['trestle', 'assemble', 'profile', '-t', '-x', 'json']
    # Tests should pass on empty set of directories.
    monkeypatch.setattr(sys, 'argv', testargs)
    rc = Trestle().run()
    assert rc == 0



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Template Versioning."""
import logging
import re
import shutil
from pathlib import Path
from typing import List, Optional, Tuple

from pkg_resources import resource_filename

from trestle.common import file_utils
from trestle.common.const import DRAWIO_FILE_EXT, MARKDOWN_FILE_EXT, TEMPLATE_VERSION_REGEX
from trestle.common.err import TrestleError
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION, TEMPLATE_VERSION_HEADER, TRESTLE_RESOURCES
from trestle.core.draw_io import DrawIO
from trestle.core.markdown.markdown_api import MarkdownAPI

logger = logging.getLogger(__name__)


class TemplateVersioning:
    """
    Template versioning solution.

    1. Load template with a specified version.
        1. Template version can be specified via -tv flag
        2. If no version specified the latest version will be used by default.
        3. Otherwise use the templates from the specified version.
    2. Backward compatibility.
        1. Version 0.0.1 will be reserved for the template versions prior to this change.
        2. If the old template path is detected (i.e. without the template version).
            then the filesystem will be updated to the new path with the version.
        3. Upon first run, old template versions prior to this change will be placed to the folder 0.0.1.
        4. If templates have no headers then 0.0.1 version will be used.
    3. Versioning organization.
        1. Template and instance version is added both: in the file system (via path) and in the headers or metadata.
    4. Instance version validation.
        1. Markdown:
            1. When validating the instance the header will be used to validate the version.
        2. Drawio:
            1. When validating the instance the metadata will be used to validate the version.
    """

    @staticmethod
    def update_template_folder_structure(task_path: Path) -> None:
        """
        Automatically detect whether the path is an old style and update it.

        An old-style path is a path of the form:
         root_folder/.trestle/author/{template_name}/{template_objects}
        The new version path is a path of the form:
         root_folder/.trestle/author/{template_name}/{version}/{template_objects}
        By default all old-style templates will be updated to version 0.0.1
        """
        TemplateVersioning._check_if_exists_and_dir(task_path)

        try:
            all_files_wo_version = list(
                filter(lambda p: p.is_file(), file_utils.iterdir_without_hidden_files(task_path))
            )

            pattern = re.compile(TEMPLATE_VERSION_REGEX)
            all_non_template_directories = list(
                filter(
                    lambda p: p.is_dir() and pattern.search(p.parts[-1]) is None,
                    file_utils.iterdir_without_hidden_files(task_path)
                )
            )

            new_dir = Path(f'{task_path}/{START_TEMPLATE_VERSION}')
            new_dir.mkdir(parents=True, exist_ok=True)

            if len(all_files_wo_version) == 0:
                logger.debug('No templates outside of the version folders.')
                # if the base template folder is empty, delete it
                if len(list(new_dir.iterdir())) == 0:
                    new_dir.rmdir()

            for f in all_files_wo_version + all_non_template_directories:
                if f.is_file():
                    shutil.copy(f, new_dir)
                elif f.is_dir():
                    shutil.copytree(f, new_dir / f.name)

            for p in all_files_wo_version + all_non_template_directories:
                if p.is_dir():
                    shutil.rmtree(p)
                else:
                    p.unlink()

        except OSError as e:
            raise TrestleError(f'Error while updating template folder: {e}')
        except Exception as e:
            raise TrestleError(f'Unexpected error while updating template folder: {e}')

    @staticmethod
    def get_versioned_template_dir(task_path: Path, version: Optional[str] = None) -> Path:
        """
        Get a template folder of the specified version.

        If no version is given, the latest version of template will be returned
        """
        TemplateVersioning._check_if_exists_and_dir(task_path)
        latest_path = None
        if version is None:
            latest_path, _ = TemplateVersioning.get_latest_version_for_task(task_path)
        else:
            latest_path = Path(f'{task_path}/{version}/')

        if not latest_path.exists():
            raise TrestleError(f'The task: {task_path} with version: {version} does not exists.')

        return latest_path

    @staticmethod
    def get_latest_version_for_task(task_path: Path) -> Tuple[Path, str]:
        """Get latest version of the template for the given task."""
        TemplateVersioning._check_if_exists_and_dir(task_path)

        all_versions = TemplateVersioning.get_all_versions_for_task(task_path)
        max_version = START_TEMPLATE_VERSION

        if len(all_versions) == 0:
            logger.debug(f'No template versions were found for task: {task_path}, defaulting to 0.0.1')
            max_version = START_TEMPLATE_VERSION
        else:
            max_version = max(all_versions)

        latest_path = Path(f'{task_path}/{max_version}')

        return latest_path, max_version

    @staticmethod
    def get_all_versions_for_task(task_path: Path) -> List[str]:
        """Get all versions for the task."""
        pattern = re.compile(TEMPLATE_VERSION_REGEX)
        all_versions = []
        max_version = START_TEMPLATE_VERSION
        for p in task_path.iterdir():
            if p.is_dir() and pattern.search(p.parts[-1]) is not None:
                match = pattern.search(p.parts[-1]).string
                all_versions.append(match)
                if match > max_version:
                    max_version = match

        return all_versions

    @staticmethod
    def write_versioned_template(
        resource_name: str, task_path: Path, target_file: Path, version: Optional[str] = None
    ) -> None:
        """
        Write a template with the header or metadata of a specified version.

        If no version was given the latest version for the task will be used.

        Args:
            resource_name:  Template resource name
            task_path: Task path
            target_file: File path where template will be written
            version: return a resource of a specific version

        Returns:
            A dotted path of a versioned template, list of all available versions
        """
        TemplateVersioning._check_if_exists_and_dir(task_path)
        try:
            templates_resource_path = TRESTLE_RESOURCES + '.templates'

            generic_template = Path(resource_filename(templates_resource_path, resource_name)).resolve()
            if version is None:
                _, version = TemplateVersioning.get_latest_version_for_task(task_path)

            # modify header/metadata in the template
            if generic_template.suffix == MARKDOWN_FILE_EXT:
                md_api = MarkdownAPI()
                header, md_body = md_api.processor.read_markdown_wo_processing(generic_template)
                header[TEMPLATE_VERSION_HEADER] = version
                md_api.write_markdown_with_header(target_file, header, md_body)
                logger.debug(f'Successfully written template markdown to {target_file}')
            elif generic_template.suffix == DRAWIO_FILE_EXT:
                drawio = DrawIO(generic_template)
                metadata = drawio.get_metadata()[0]
                metadata[TEMPLATE_VERSION_HEADER] = version

                drawio.write_drawio_with_metadata(generic_template, metadata, 0, target_file)
                logger.debug(f'Successfully written template drawio to {target_file}')
            else:
                raise TrestleError(f'Unsupported template file extension {generic_template.suffix}')
        except OSError as e:
            raise TrestleError(f'Error while updating template folder: {e}')

    @staticmethod
    def is_valid_version(template_version: str) -> bool:
        """Check if the version format is correct."""
        if template_version is None:
            return True  # we can have empty version
        if template_version == '0.0.0':
            return False
        version_regex = r'^[0-9]+.[0-9]+.[0-9]+$'
        pattern = re.compile(version_regex)
        if pattern.search(template_version):
            return True
        else:
            return False

    @staticmethod
    def _check_if_exists_and_dir(task_path: Path) -> None:
        if not task_path.exists():
            raise TrestleError(f'Path: {task_path} does not exists.')

        if not task_path.is_dir():
            raise TrestleError(f'File {task_path} passed, however template directory is expected.')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validate based on all registered validators."""

import trestle.core.validator_factory as vfact
from trestle.core.base_model import OscalBaseModel
from trestle.core.validator import Validator


class AllValidator(Validator):
    """Validator to confirm the model passes all registered validation tests."""

    last_failure_msg: str

    def error_msg(self) -> str:
        """Return information on which validation failed."""
        return self.last_failure_msg

    def model_is_valid(self, model: OscalBaseModel, quiet: bool) -> bool:
        """
        Validate an oscal model against all available validators in the trestle library.

        args:
            model: An Oscal model that can be passed to the validator.
            quiet: Don't report msgs unless invalid.

        returns:
            True (valid) if the model passed all registered validators.
        """
        self.last_failure_msg = self.__doc__
        for val in vfact.validator_factory.get_all():
            if val != self and not val.model_is_valid(model, quiet):
                self.last_failure_msg = val.error_msg()
                return False
        return True



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the profile author module."""

import argparse
import pathlib
import shutil
import sys
from typing import Dict, Optional, Tuple

from _pytest.monkeypatch import MonkeyPatch

import pytest

from ruamel.yaml import YAML

from tests import test_utils

import trestle.common.const as const
import trestle.oscal.catalog as cat
import trestle.oscal.common as com
import trestle.oscal.profile as prof
from trestle.cli import Trestle
from trestle.common import file_utils
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog_interface import CatalogInterface
from trestle.core.commands.author.profile import ProfileAssemble, ProfileGenerate, sections_to_dict
from trestle.core.control_interface import ControlInterface
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.markdown.markdown_node import MarkdownNode
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver

# test dicts are of form {'name_exp': [(name, exp_str)...], 'text': prose}
# the text is appended to the end of the file
# then the assembled control is searched for exp_str in the prose of the named parts

markdown_name = 'my_md'
prof_name = 'my_prof'
md_name = 'my_md'
assembled_prof_name = 'my_assembled_prof'

my_guidance_text = """
## Control My Guidance

This is My Guidance.
"""

# name_exp maps the name of the section to the expected prose
# text is the text to be inserted in the markdown at the end of the file

# just add a new addition
my_guidance_dict = {'name_exp': [('ac-1_my_guidance', 'my_guidance', 'This is My Guidance.')], 'text': my_guidance_text}

multi_guidance_text = """
## Control A Guidance

This is A Guidance.

## Control B Guidance

This is B Guidance.
"""
# add two additions
multi_guidance_dict = {
    'name_exp': [
        ('ac-1_a_guidance', 'a_guidance', 'This is A Guidance.'),
        ('ac-1_b_guidance', 'b_guidance', 'This is B Guidance.')
    ],
    'text': multi_guidance_text
}

control_subparts_text = """
## Control A Guidance

Control A prose

### A Subpart

A subpart prose

#### A Subsubpart

A subsubpart prose

### B Subpart

B subpart prose

## Part a.

### a by_id subpart

a by_id subpart prose
"""

# part.id, part.name, part.prose
control_subparts_dict = {
    'name_exp': [
        ('ac-1_a_guidance', 'a_guidance', 'Control A prose'),
        ('ac-1_a_guidance.a_subpart', 'a_subpart', 'A subpart prose'),
        ('ac-1_a_guidance.a_subpart.a_subsubpart', 'a_subsubpart', 'A subsubpart prose'),
        ('ac-1_a_guidance.b_subpart', 'b_subpart', 'B subpart prose'),
        ('ac-1_smt.a', 'item', 'Develop, document, and disseminate'),  # this is the original NIST prose
        ('ac-1_smt.a.a_by_id_subpart', 'a_by_id_subpart', 'a by_id subpart prose')
    ],
    'text': control_subparts_text
}

all_sections_str = (
    'guidance:Guidance,implgdn:Implementation Guidance,expevid:Expected Evidence,my_guidance:My Guidance,'
    'a_guidance:A Guidance,b_guidance:B Guidance,NeededExtra:Needed Extra,a_subpart:A Subpart,'
    'a_subsubpart:A Subsubpart,b_subpart:B Subpart,a_by_id_subpart:a by_id subpart'
)

all_sections_dict = {
    'guidance': 'Guidance',
    'implgdn': 'Implementation Guidance',
    'expevid': 'Expected Evidence',
    'my_guidance': 'My Guidance',
    'a_guidance': 'A Guidance',
    'b_guidance': 'B Guidance',
    'NeededExtra': 'Needed Extra',
    'a_subpart': 'A Subpart',
    'a_subsubpart': 'A Subsubpart',
    'b_subpart': 'B Subpart',
    'a_by_id_subpart': 'a by_id subpart'
}


def edit_files(control_path: pathlib.Path, change_parameters: bool, guid_dict: Dict[str, str]) -> None:
    """Edit the files to show assemble worked."""
    assert control_path.exists()
    assert file_utils.insert_text_in_file(control_path, None, guid_dict['text'])
    if control_path.stem == 'ac-1':
        assert test_utils.replace_line_in_file_after_tag(
            control_path, 'prop with ns', '    ns: https://my_new_namespace\n'
        )
        assert file_utils.insert_text_in_file(control_path, 'prop with no ns', '    ns: https://my_added_namespace\n')
    if change_parameters:
        # delete profile values for 4, then replace value for 3 with new value
        assert test_utils.substitute_text_in_file(control_path, 'officer', 'new value')
        assert test_utils.delete_line_in_file(control_path, 'weekly')
        assert test_utils.delete_line_in_file(control_path, 'label:')
        assert file_utils.insert_text_in_file(control_path, 'ac-1_prm_1:', '    label: label from edit\n')


def setup_profile_generate(trestle_root: pathlib.Path,
                           source_prof_name: str) -> Tuple[pathlib.Path, pathlib.Path, pathlib.Path, pathlib.Path]:
    """Set up files for profile generate."""
    nist_catalog_path = test_utils.JSON_TEST_DATA_PATH / test_utils.SIMPLIFIED_NIST_CATALOG_NAME
    trestle_cat_dir = trestle_root / 'catalogs/nist_cat'
    trestle_cat_dir.mkdir(exist_ok=True, parents=True)
    shutil.copy(nist_catalog_path, trestle_cat_dir / 'catalog.json')
    profile_dir = trestle_root / f'profiles/{prof_name}'
    profile_dir.mkdir(parents=True, exist_ok=True)
    # simple test profile sets values for ac-1 params 1-6 but not param_7
    source_prof_path = test_utils.JSON_TEST_DATA_PATH / source_prof_name
    profile_path = profile_dir / 'profile.json'
    shutil.copy(source_prof_path, profile_path)
    markdown_path = trestle_root / md_name
    ac1_path = markdown_path / 'ac/ac-1.md'
    assembled_prof_dir = trestle_root / f'profiles/{assembled_prof_name}'
    return ac1_path, assembled_prof_dir, profile_path, markdown_path


@pytest.mark.parametrize('add_header', [True, False])
@pytest.mark.parametrize('guid_dict', [my_guidance_dict, multi_guidance_dict, control_subparts_dict])
@pytest.mark.parametrize('use_cli', [True, False])
@pytest.mark.parametrize('dir_exists', [True, False])
@pytest.mark.parametrize('set_parameters_flag', [True, False])
def test_profile_generate_assemble(
    add_header: bool,
    guid_dict: Dict,
    use_cli: bool,
    dir_exists: bool,
    set_parameters_flag: bool,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Test the profile markdown generator."""
    ac1_path, assembled_prof_dir, profile_path, markdown_path = setup_profile_generate(
        tmp_trestle_dir,
        'simple_test_profile.json'
    )
    yaml_header_path = test_utils.YAML_TEST_DATA_PATH / 'good_simple.yaml'

    ac_path = markdown_path / 'ac'

    # convert resolved profile catalog to markdown then assemble it after adding an item to a control
    # generate, edit, assemble
    if use_cli:
        test_args = f'trestle author profile-generate -n {prof_name} -o {md_name} -rs NeededExtra'.split(  # noqa E501
        )
        if add_header:
            test_args.extend(['-y', str(yaml_header_path)])
        test_args.extend(['-s', all_sections_str])
        monkeypatch.setattr(sys, 'argv', test_args)

        assert Trestle().run() == 0

        fc = test_utils.FileChecker(ac_path)

        assert Trestle().run() == 0

        assert fc.files_unchanged()

        edit_files(ac1_path, True, guid_dict)

        test_args = f'trestle author profile-assemble -n {prof_name} -m {md_name} -o {assembled_prof_name}'.split()
        if set_parameters_flag:
            test_args.append('-sp')
        if dir_exists:
            assembled_prof_dir.mkdir()
        monkeypatch.setattr(sys, 'argv', test_args)
        assert Trestle().run() == 0
    else:
        profile_generate = ProfileGenerate()
        yaml_header = {}
        if add_header:
            yaml = YAML()
            yaml_header = yaml.load(yaml_header_path.open('r'))
        sections_dict = sections_to_dict(all_sections_str)

        profile_generate.generate_markdown(
            tmp_trestle_dir, profile_path, markdown_path, yaml_header, False, sections_dict, 'NeededExtra'
        )

        fc = test_utils.FileChecker(ac_path)

        profile_generate.generate_markdown(
            tmp_trestle_dir, profile_path, markdown_path, yaml_header, False, sections_dict, 'NeededExtra'
        )

        assert fc.files_unchanged()

        edit_files(ac1_path, True, guid_dict)

        if dir_exists:
            assembled_prof_dir.mkdir()
        assert ProfileAssemble.assemble_profile(
            tmp_trestle_dir,
            prof_name,
            md_name,
            assembled_prof_name,
            set_parameters_flag,
            False,
            None,
            None,
            None,
            None
        ) == 0

    assert test_utils.confirm_text_in_file(ac1_path, const.TRESTLE_GLOBAL_TAG, 'title: Trestle test profile')

    # check the assembled profile is as expected
    profile: prof.Profile
    profile, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, assembled_prof_name,
                                                 prof.Profile, FileContentType.JSON)
    assert ModelUtils.model_age(profile) < test_utils.NEW_MODEL_AGE_SECONDS
    # get the set_params from the assembled profile
    set_params = profile.modify.set_parameters
    if set_parameters_flag:
        assert set_params[2].values[0].__root__ == 'new value'
        assert set_params[1].props[0].ns == const.TRESTLE_GENERIC_NS
        assert len(set_params) == 14
    else:
        # the original profile did not have ns set for this display name
        # confirm the namespace is not defined unless set_parameters_flag is True
        # i.e. the setting of ns for display-name is not automatic unless set-parameters is true
        assert set_params[2].values[0].__root__ == 'officer'
        assert set_params[1].props[0].ns is None
        assert len(set_params) == 15
    assert set_params[0].param_id == 'ac-1_prm_1'
    assert set_params[0].values[0].__root__ == 'all personnel'
    assert set_params[0].props[0].name == const.DISPLAY_NAME
    assert set_params[0].props[0].value.startswith('Pretty')
    assert set_params[0].props[0].ns == const.TRESTLE_GENERIC_NS
    assert set_params[1].param_id == 'ac-1_prm_2'
    assert set_params[1].values[0].__root__ == 'Organization-level'
    assert set_params[1].values[1].__root__ == 'System-level'
    assert set_params[1].props[0].name == const.DISPLAY_NAME
    assert set_params[2].param_id == 'ac-1_prm_3'
    add = profile.modify.alters[0].adds[0]
    assert add.props[0].ns == 'https://my_new_namespace'
    assert add.props[1].ns == 'https://my_added_namespace'

    # now create the resolved profile catalog from the assembled json profile and confirm the addition is there

    assem_prof_path = assembled_prof_dir / 'profile.json'
    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, assem_prof_path)
    catalog_interface = CatalogInterface(catalog)
    # confirm correct ids, names, and prose for the parts
    ac_1 = catalog_interface.get_control('ac-1')
    for part_id, name, exp_str in guid_dict['name_exp']:
        part = ControlInterface.get_part_by_id(ac_1, part_id)
        assert part.name == name
        assert part.prose.find(exp_str) >= 0

    # regen and make sure markdown is unchanged

    # ac_2.1 has changes to the comment section due to parts not being present, then being added - so remove it
    ac_21_path = tmp_trestle_dir / 'my_md/ac/ac-2.1.md'
    ac_21_path.unlink()

    fc = test_utils.FileChecker(ac_path)

    profile_generate = ProfileGenerate()
    profile_generate.generate_markdown(
        tmp_trestle_dir, assem_prof_path, markdown_path, {}, False, all_sections_dict, 'NeededExtra'
    )

    # if not set_parameters_flag then the markdown will have no profile value for ac-1_prm_4, but the profile will still
    # have a value for it.  So if set_parameters_flag is true we need to remove that line that gets added
    if not set_parameters_flag:
        ac_1_path = tmp_trestle_dir / 'my_md/ac/ac-1.md'
        assert test_utils.delete_line_in_file(ac_1_path, 'profile-values: weekly')

    ac_21_path.unlink()
    assert fc.files_unchanged()


@pytest.mark.parametrize(
    'required_sections, success', [(None, True), ('a_guidance,b_guidance', True), ('a_guidance,c_guidance', False)]
)
@pytest.mark.parametrize('ohv', [True, False])
def test_profile_ohv(required_sections: Optional[str], success: bool, ohv: bool, tmp_trestle_dir: pathlib.Path) -> None:
    """Test profile generate assemble with overwrite-header-values."""
    ac1_path, assembled_prof_dir, profile_path, markdown_path = setup_profile_generate(
        tmp_trestle_dir,
        'simple_test_profile.json'
    )
    yaml_header_path = test_utils.YAML_TEST_DATA_PATH / 'good_simple.yaml'
    new_version = '1.2.3'

    # convert resolved profile catalog to markdown then assemble it after adding an item to a control
    # if set_parameters_flag is true, the yaml header will contain all the parameters
    profile_generate = ProfileGenerate()
    yaml = YAML()
    yaml_header = yaml.load(yaml_header_path.open('r'))
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, yaml_header, ohv, None, None)

    edit_files(ac1_path, True, multi_guidance_dict)
    markdown_path = tmp_trestle_dir / md_name
    # change guidance in the other two controls but don't change header
    ac2_path = markdown_path / 'ac/ac-2.md'
    ac21_path = markdown_path / 'ac/ac-2.1.md'
    edit_files(ac2_path, False, multi_guidance_dict)
    edit_files(ac21_path, False, multi_guidance_dict)

    if success:
        # always doing set_params True
        assert ProfileAssemble.assemble_profile(
            tmp_trestle_dir,
            prof_name,
            md_name,
            assembled_prof_name,
            True,
            False,
            new_version,
            None,
            required_sections,
            None
        ) == 0

        # check the assembled profile is as expected
        profile: prof.Profile
        profile, _ = ModelUtils.load_top_level_model(
            tmp_trestle_dir, assembled_prof_name,
            prof.Profile,
            FileContentType.JSON
        )
        set_params = profile.modify.set_parameters

        assert len(set_params) == 14
        assert set_params[0].values[0].__root__ == 'all personnel'
        # the label is present in the header so it ends up in the set_parameter
        assert set_params[0].label == 'label from edit'
        assert set_params[1].param_id == 'ac-1_prm_2'
        assert set_params[1].values[0].__root__ == 'Organization-level'
        assert set_params[1].values[1].__root__ == 'System-level'
        assert set_params[2].values[0].__root__ == 'new value'
        assert profile.metadata.version.__root__ == new_version
        if ohv:
            assert set_params[3].values[0].__root__ == 'no meetings'
            assert set_params[3].label == 'meetings cancelled'
        else:
            assert set_params[3].values[0].__root__ == 'all meetings'
            assert set_params[3].label is None

        catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, assembled_prof_dir / 'profile.json')
        catalog_interface = CatalogInterface(catalog)
        # confirm correct ids, names, and prose for the parts
        ac_1 = catalog_interface.get_control('ac-1')
        for part_id, name, exp_str in multi_guidance_dict['name_exp']:
            part = ControlInterface.get_part_by_id(ac_1, part_id)
            assert part.name == name
            assert part.prose.find(exp_str) >= 0

    else:
        with pytest.raises(TrestleError):
            ProfileAssemble.assemble_profile(
                tmp_trestle_dir,
                prof_name,
                md_name,
                assembled_prof_name,
                True,
                False,
                new_version,
                None,
                required_sections,
                None
            )


def test_profile_failures(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test failure modes of profile generate and assemble."""
    # no trestle root specified direct command
    test_args = argparse.Namespace(name='my_prof', output='new_prof', verbose=0, set_parameters=False, sections=None)
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 1

    # no trestle root specified
    profile_assemble = ProfileAssemble()
    assert profile_assemble._run(test_args) == 1

    # bad yaml
    bad_yaml_path = str(test_utils.YAML_TEST_DATA_PATH / 'bad_simple.yaml')
    trestle_root = tmp_trestle_dir
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name='my_prof',
        output='new_prof',
        yaml_header=bad_yaml_path,
        verbose=0,
        set_parameters=False,
        sections=None
    )
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 1

    # profile not available for load
    test_args = 'trestle author profile-generate -n my_prof -o my_md -v'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1

    # setup for generate and assemble
    profile_path = ModelUtils.path_for_top_level_model(tmp_trestle_dir, 'my_prof', prof.Profile, FileContentType.JSON)
    profile_path.parent.mkdir()
    shutil.copyfile(test_utils.JSON_TEST_DATA_PATH / 'simple_test_profile.json', profile_path)
    cat_path = ModelUtils.path_for_top_level_model(tmp_trestle_dir, 'nist_cat', cat.Catalog, FileContentType.JSON)
    cat_path.parent.mkdir()
    shutil.copyfile(test_utils.JSON_NIST_DATA_PATH / test_utils.JSON_NIST_CATALOG_NAME, cat_path)

    # generate markdown with required section but don't give it any prose
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name='my_prof',
        output='md_prof',
        verbose=0,
        set_parameters_flag=False,
        overwrite_header_values=False,
        yaml_header=None,
        sections='NeededExtra:Needed Extra,implgdn:Implementation Guidance,expevid:Expected Evidence',
        required_sections='NeededExtra'
    )
    profile_generate = ProfileGenerate()
    assert profile_generate._run(test_args) == 0

    # set arguments for assemble
    test_args = argparse.Namespace(
        trestle_root=trestle_root,
        name=None,
        markdown='md_prof',
        output='my_prof',
        verbose=0,
        set_parameters=False,
        yaml_header=None,
        required_sections='NeededExtra',
        regenerate=False,
        version=None,
        allowed_sections=None,
        sections=None
    )
    # fail since required section not filled in
    profile_assemble = ProfileAssemble()
    assert profile_assemble._run(test_args) == 1

    # succeed if not specifying required section
    test_args.required_sections = None
    assert profile_assemble._run(test_args) == 0

    # fail if allowed sections doesn't include all needed
    test_args.allowed_sections = 'NeededExtra,DummySection'
    assert profile_assemble._run(test_args) == 1

    # succed if allowed sections has all
    test_args.allowed_sections = 'expevid,implgdn,NeededExtra'
    assert profile_assemble._run(test_args) == 0

    # disallowed output name
    test_args = 'trestle author profile-generate -n my_prof -o profiles -v'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    assert Trestle().run() == 1


def test_profile_overwrite(tmp_trestle_dir: pathlib.Path) -> None:
    """Test blocking overwrite if no change to assembled profile relative to one it would overwrite."""
    _, _, profile_path, markdown_path = setup_profile_generate(tmp_trestle_dir, 'simple_test_profile.json')

    # generate the markdown and assemble
    profile_generate = ProfileGenerate()
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, all_sections_dict, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None, None
    ) == 0

    # note the file timestamp, then regenerate and assemble again
    orig_time = profile_path.stat().st_mtime

    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, all_sections_dict, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None, None
    ) == 0

    # the timestamp should be the same, indicating the file was not written
    new_time = profile_path.stat().st_mtime

    assert new_time == orig_time

    # now generate again but with different section title, causing change in generated profile markdown
    new_sections = {'implgdn': 'Different Title'}
    profile_generate.generate_markdown(tmp_trestle_dir, profile_path, markdown_path, {}, False, new_sections, None)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, prof_name, True, False, None, None, None, None
    ) == 0

    # the timestamp should now be different
    new_time = profile_path.stat().st_mtime

    assert new_time != orig_time


def test_profile_alter_adds(simplified_nist_profile: prof.Profile) -> None:
    """Test profile alter adds involving Nones."""
    adds = [prof.Add(props=[com.Property(name='prop_1', value='prop_1_val')])]
    alters = [prof.Alter(control_id='ac-1', adds=adds)]
    # first case modify is None
    assert ProfileAssemble._replace_alter_adds(simplified_nist_profile, alters)
    simplified_nist_profile.modify.alters = None
    # second case modify is not None but alters is None
    assert ProfileAssemble._replace_alter_adds(simplified_nist_profile, alters)


def _check_parts(part: com.Part, prose_1: str) -> None:
    assert part.parts[0].id == 'ac-1_smt.b.new_guidance'
    assert part.parts[0].name == 'new_guidance'
    assert part.parts[0].prose == 'This is my added prose for a part in the statement'
    assert part.parts[1].id == 'ac-1_smt.b.new_evidence'
    assert part.parts[1].prose == prose_1


def _check_multi_section(part: com.Part) -> None:
    assert part.id == 'ac-1_multi_section'
    assert part.parts[0].id == 'ac-1_multi_section.sub_a'
    assert part.parts[0].parts[0].id == 'ac-1_multi_section.sub_a.sub_sub_a'


def test_profile_alter_props(tmp_trestle_dir: pathlib.Path) -> None:
    """Test profile alter adds involving props."""
    ac1_path, _, profile_path, markdown_path = setup_profile_generate(tmp_trestle_dir, 'profile_with_alter_props.json')
    sections = {'implgdn': 'Implementation Guidance', 'expevid': 'Expected Evidence', 'guidance': 'Guidance'}
    # generate markdown twice and confirm no changes
    profile_generate = ProfileGenerate()
    assert profile_generate.generate_markdown(
        tmp_trestle_dir, profile_path, markdown_path, {}, False, sections, None
    ) == 0

    fc = test_utils.FileChecker(ac1_path.parent)

    assert profile_generate.generate_markdown(
        tmp_trestle_dir, profile_path, markdown_path, {}, False, sections, None
    ) == 0

    assert fc.files_unchanged()

    text = """  - name: ac-1_new
    value: ac-1 new value
    remarks: ac-1 new stuff
  - name: ac-1_new_part
    value: ac-1 new part value
    smt-part: c.
  - name: prop_in_expevid
    value: value in expevid
    smt-part: ac-1_expevid
  - name: cannot_add_to_resolved_cat_because_no_part_id
    value: doesnt work
    smt-part: foo_bar
"""

    # insert four new props, one is by id attached to part c., another is to ac-1_expevid
    # a final one is attached to cannot_add and will give warning that it isn't found on assemble
    assert file_utils.insert_text_in_file(ac1_path, const.TRESTLE_ADD_PROPS_TAG, text)

    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, assembled_prof_name, True, False, None, sections, None, None
    ) == 0

    # check the assembled profile is as expected
    profile: prof.Profile
    profile, prof_path = ModelUtils.load_top_level_model(
        tmp_trestle_dir,
        assembled_prof_name,
        prof.Profile, FileContentType.JSON
    )
    adds = profile.modify.alters[0].adds
    assert len(adds) == 5
    assert adds[0].position == prof.Position.ending
    assert adds[0].by_id is None
    assert len(adds[0].parts) == 2
    assert len(adds[0].props) == 2
    assert adds[1].by_id == 'ac-1_expevid'
    assert adds[1].props[0].name == 'prop_in_expevid'
    assert adds[1].props[0].value == 'value in expevid'
    assert adds[2].by_id == 'ac-1_smt.a'
    assert adds[3].by_id == 'ac-1_smt.c'
    assert adds[4].by_id == 'foo_bar'

    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, prof_path)
    ac1 = catalog.groups[0].controls[0]
    assert ac1.parts[0].parts[0].id == 'ac-1_smt.a'
    assert ac1.parts[0].parts[0].props[1].name == 'ac1_a_foo'
    assert ac1.parts[0].parts[0].props[1].value == 'ac1 a bar'
    assert ac1.parts[0].parts[2].id == 'ac-1_smt.c'
    assert ac1.parts[0].parts[2].props[1].name == 'ac-1_new_part'
    assert ac1.parts[0].parts[2].props[1].value == 'ac-1 new part value'
    assert ac1.parts[3].props[0].name == 'prop_in_expevid'

    prose = """
## Control Multi Section

### Sub A

Text in Sub A

#### Sub Sub A

Text in Sub Sub A

### Sub B

Text in Sub B

## Part b.

### New Guidance

This is my added prose for a part in the statement

### New Evidence

More evidence

"""
    sections['sub_a'] = 'Sub A'
    sections['sub_sub_a'] = 'Sub Sub A'
    sections['sub_b'] = 'Sub B'
    sections['newguidance'] = 'New Guidance'
    sections['newevidence'] = 'New Evidence'

    assert file_utils.insert_text_in_file(ac1_path, None, prose)
    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, assembled_prof_name, True, False, None, sections, None, None
    ) == 0

    # check the assembled profile is as expected
    profile, prof_path = ModelUtils.load_top_level_model(
        tmp_trestle_dir,
        assembled_prof_name,
        prof.Profile, FileContentType.JSON
    )
    adds = profile.modify.alters[0].adds
    assert len(adds) == 6
    assert adds[0].position == prof.Position.ending
    assert adds[1].by_id == 'ac-1_expevid'
    assert adds[2].by_id == 'ac-1_smt.a'
    assert adds[3].by_id == 'ac-1_smt.b'
    assert adds[4].by_id == 'ac-1_smt.c'
    assert adds[0].parts[2].id == 'ac-1_multi_section'
    assert adds[0].parts[2].parts[0].id == 'ac-1_multi_section.sub_a'
    assert adds[0].parts[2].parts[0].parts[0].id == 'ac-1_multi_section.sub_a.sub_sub_a'

    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, prof_path)
    _check_parts(catalog.groups[0].controls[0].parts[0].parts[1], 'More evidence')

    # Confirm that changed prose in the markdown is retained on new profile-generate
    assert test_utils.substitute_text_in_file(ac1_path, 'More evidence', 'Updated evidence')
    assert profile_generate.generate_markdown(tmp_trestle_dir, prof_path, markdown_path, {}, False, sections, None) == 0
    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, md_name, assembled_prof_name, True, False, None, sections, None, None
    ) == 0
    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, prof_path)
    _check_parts(catalog.groups[0].controls[0].parts[0].parts[1], 'Updated evidence')
    _check_multi_section(part=catalog.groups[0].controls[0].parts[4])

    fresh_md_name = 'fresh_markdown'
    fresh_md_path = tmp_trestle_dir / fresh_md_name
    fresh_assem_prof_name = 'fresh_assem'
    fresh_assem_prof_path = tmp_trestle_dir / 'profiles' / fresh_assem_prof_name / 'profile.json'
    assert profile_generate.generate_markdown(tmp_trestle_dir, prof_path, fresh_md_path, {}, False, sections, None) == 0
    assert ProfileAssemble.assemble_profile(
        tmp_trestle_dir, prof_name, fresh_md_name, fresh_assem_prof_name, True, False, None, sections, None, None
    ) == 0
    catalog = ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, fresh_assem_prof_path)
    _check_parts(catalog.groups[0].controls[0].parts[0].parts[1], 'Updated evidence')
    _check_multi_section(part=catalog.groups[0].controls[0].parts[4])


def test_adding_removing_sections(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test the profile generate and assemble in cycles, with incremental changes."""

    def generate_assemble(md_path: pathlib.Path) -> MarkdownNode:
        prof_assemble = f'trestle author profile-assemble -n main_profile -m {md_name} -o main_profile'
        test_utils.execute_command_and_assert(prof_assemble, 0, monkeypatch)

        # generate markdown again, ensure section is there
        prof_generate = f'trestle author profile-generate -n main_profile -o {md_name}'
        test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)

        assert md_path.exists()
        _, tree = md_api.processor.process_markdown(md_path)

        return tree

    test_utils.setup_for_multi_profile(tmp_trestle_dir, True, True)

    prof_generate = f'trestle author profile-generate -n main_profile -o {md_name}'
    test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)

    ac1_path = tmp_trestle_dir / md_name / 'ac/ac-1.md'
    ac2_path = tmp_trestle_dir / md_name / 'ac/ac-2.md'

    md_api = MarkdownAPI()

    assert ac1_path.exists()
    assert file_utils.insert_text_in_file(ac1_path, None, '## Control this_should_appear_in_parts \n Test text.')
    _, tree = md_api.processor.process_markdown(ac1_path)

    assert tree.get_node_for_key('## Control this_should_appear_in_parts', strict_matching=True)

    # Scenario 1: Profile has no alters at all, ensure new sections is added
    tree = generate_assemble(ac1_path)
    assert tree.get_node_for_key('## Control this_should_appear_in_parts', strict_matching=True)

    # Scenario 2: Profiles has one alter for ac-1, now add new section to ac-2, ensure new sectio is added
    assert ac2_path.exists()
    assert file_utils.insert_text_in_file(ac2_path, None, '## Control this_should_appear_in_parts2 \n Test text.')
    _, tree = md_api.processor.process_markdown(ac2_path)
    assert tree.get_node_for_key('## Control this_should_appear_in_parts2', strict_matching=True)

    tree = generate_assemble(ac2_path)
    assert tree.get_node_for_key('## Control this_should_appear_in_parts2', strict_matching=True)

    # Scenario 3: Remove added section from the ac-1, ensure the alters are deleted
    assert ac1_path.exists()
    assert test_utils.delete_line_in_file(ac1_path, '## Control this_should_appear_in_parts', 2)
    _, tree = md_api.processor.process_markdown(ac1_path)
    assert not tree.get_node_for_key('## Control this_should_appear_in_parts')

    tree = generate_assemble(ac1_path)
    assert not tree.get_node_for_key('## Control this_should_appear_in_parts')


@pytest.mark.parametrize('bracket_format', [True, False])
@pytest.mark.parametrize('show_values', [True, False])
def test_profile_resolve(
    tmp_trestle_dir: pathlib.Path, show_values: bool, bracket_format: bool, monkeypatch: MonkeyPatch
) -> None:
    """Test profile resolve to create resolved profile catalog."""
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, False)
    cat_name = 'resolved_catalog'
    command_profile_resolve = f'trestle author profile-resolve -n main_profile -o {cat_name}'
    if show_values:
        command_profile_resolve += ' -sv'
    if bracket_format:
        command_profile_resolve += ' -bf [(.])'
    test_utils.execute_command_and_assert(command_profile_resolve, 0, monkeypatch)
    res_cat, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, cat_name, cat.Catalog, FileContentType.JSON)
    ac_1 = res_cat.groups[0].controls[0]
    expected_value = '{{ insert: param, ac-1_prm_3 }}'
    if show_values:
        expected_value = 'officer'
        if bracket_format:
            expected_value = f'[({expected_value}])'
    expected_prose = f'Designate an {expected_value} to manage the development, documentation, and dissemination of the access control policy and procedures; and'  # noqa E501
    assert ac_1.parts[0].parts[1].prose == expected_prose


@pytest.mark.parametrize('show_values', [True, False])
def test_profile_resolve_fail(tmp_trestle_dir: pathlib.Path, show_values: bool, monkeypatch: MonkeyPatch) -> None:
    """Test profile resolve to create resolved profile catalog."""
    # confirm failure for non-existent profile
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, False)
    cat_name = 'resolved_catalog'
    command_profile_resolve = f'trestle author profile-resolve -n foo -o {cat_name}'
    if show_values:
        command_profile_resolve += ' -sv'
    test_utils.execute_command_and_assert(command_profile_resolve, 1, monkeypatch)

    # confirm failure for existing but corrupt profile
    bad_prof_dir = tmp_trestle_dir / 'profiles/bad_prof'
    bad_prof_dir.mkdir(exist_ok=True, parents=True)
    src_path = test_utils.JSON_TEST_DATA_PATH / 'bad_simple.json'
    shutil.copy2(str(src_path), str(bad_prof_dir / 'profile.json'))
    command_profile_resolve = f'trestle author profile-resolve -n bad_prof -o {cat_name}'
    if show_values:
        command_profile_resolve += ' -sv'
    test_utils.execute_command_and_assert(command_profile_resolve, 1, monkeypatch)


def test_profile_generate_updates_statement(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test proper update of statement contents on regenerate."""
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, True)
    prof_generate = f'trestle author profile-generate -n main_profile -o {md_name} -ohv'
    test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)

    # confirm that disallowed edits to the control statement are overwritten on next generate
    bad_text = '  - \[3.\] Count kangaroos.'
    ac1_path = tmp_trestle_dir / md_name / 'ac/ac-1.md'
    tag = 'param, ac-1_prm_7'
    file_utils.insert_text_in_file(ac1_path, tag, bad_text)
    assert test_utils.confirm_text_in_file(ac1_path, tag, bad_text)
    test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)
    assert not test_utils.confirm_text_in_file(ac1_path, tag, bad_text)

    # put bad text back in disallowed part of control statement
    file_utils.insert_text_in_file(ac1_path, tag, bad_text)

    # add a new allowed part to the profile
    # and change a parameter value on a control
    main_prof, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, 'main_profile', prof.Profile, FileContentType.JSON)
    main_prof.modify.alters[0].adds[0].parts.append(com.Part(id='ac-1_wombat', name='wombat', prose='Assess wombats.'))
    main_prof.modify.set_parameters[2].values = [com.ParameterValue(__root__='echidna')]
    ModelUtils.save_top_level_model(main_prof, tmp_trestle_dir, 'main_profile', FileContentType.JSON)

    # now add a part to the catalog
    nist_cat, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, 'nist_cat', cat.Catalog, FileContentType.JSON)
    nist_cat.groups[0].controls[0].parts.append(com.Part(id='ac-1_koala', name='koala', prose='Enjoy koalas'))
    ModelUtils.save_top_level_model(nist_cat, tmp_trestle_dir, 'nist_cat', FileContentType.JSON)

    # generate the markdown from the profile
    prof_generate = f'trestle author profile-generate -n main_profile -o {md_name} -ohv'
    test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)

    # assemble into new profile
    prof_assem = f'trestle author profile-assemble -n main_profile -o assem_prof -m {md_name} -sp'
    test_utils.execute_command_and_assert(prof_assem, 0, monkeypatch)

    prof_resolve = 'trestle author profile-resolve -n main_profile -o resolved_cat -sv'
    test_utils.execute_command_and_assert(prof_resolve, 0, monkeypatch)

    # echidna and wombat came from changes to main profile
    # koala came from edit to the nist catalog
    resolved_cat, _ = ModelUtils.load_top_level_model(
        tmp_trestle_dir,
        'resolved_cat',
        cat.Catalog,
        FileContentType.JSON
    )
    ac1 = resolved_cat.groups[0].controls[0]
    assert ac1.params[2].values[0].__root__ == 'echidna'
    assert ac1.parts[3].id == 'ac-1_wombat'
    assert ac1.parts[5].id == 'ac-1_koala'


def test_profile_generate_inherited_props(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test generation of inherited props in header."""
    test_utils.setup_for_multi_profile(tmp_trestle_dir, False, True)
    prof_generate = f'trestle author profile-generate -n test_profile_f -o {md_name} -ohv'
    test_utils.execute_command_and_assert(prof_generate, 0, monkeypatch)

    md_path = tmp_trestle_dir / 'my_md/ac/ac-3.3.md'
    assert md_path.exists()
    md_api = MarkdownAPI()
    header, _ = md_api.processor.process_markdown(md_path)
    inherited_props = header[const.TRESTLE_INHERITED_PROPS_TAG]
    assert len(inherited_props) == 2
    assert inherited_props[0] == {'name': 'add_prof_b_prop', 'value': 'add prof b prop value'}
    assert inherited_props[1] == {
        'name': 'add_prof_b_prop_by_id', 'value': 'add prof b prop by id value', 'part_name': 'ac-3.3_prm_2'
    }
    set_params = header[const.SET_PARAMS_TAG]
    assert set_params['ac-3.3_prm_1'][const.PROFILE_VALUES] == 'from prof f set-param'
    assert set_params['ac-3.3_prm_1'][const.VALUES] == 'key power users'

    ac5_path = tmp_trestle_dir / 'my_md/ac/ac-5.md'
    assert test_utils.confirm_text_in_file(ac5_path, 'value: one', 'smt-part: a.')
    assert test_utils.confirm_text_in_file(ac5_path, 'test_five', 'smt-part: ac-5_gdn')
    assert test_utils.confirm_text_in_file(ac5_path, 'value: two', 'ns: https://prof_f_ns')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle custom jinja functionality."""

import pathlib
from datetime import date

from jinja2 import Environment, FileSystemLoader

import pytest

import trestle.core.jinja as tres_jinja
from trestle.core.markdown import markdown_const

JINJA_MD = 'jinja_markdown_include'


@pytest.mark.parametrize(
    'template, expected, expect_exception',
    [
        ('MDSection_include_top_level.jinja.md', 'MDSection_include_top_level.md', False),
        ('MDSection_include_top_level_adjusted.jinja.md', 'MDSection_include_top_level_adjusted.md', False),
        ('MDSection_include_nested.jinja.md', 'MDSection_include_nested.md', False),
        ('env_nested_c_c.jinja.md', 'env_nested_c_c.md', False),
        ('env_nested_c_n.jinja.md', 'env_nested_c_n.md', False), ('env_include.jinja.md', 'env_include.md', False),
        ('env_include_adjusted.jinja.md', 'env_include_adjusted.md', False),
        ('env_nested_c_c_double.jinja.md', 'env_nested_c_c_double.md', False)
    ]
)
def test_jinja_md(testdata_dir: pathlib.Path, template: str, expected: str, expect_exception: bool) -> None:
    """Test jinja markdown codes."""
    jinja_md_dir = testdata_dir / JINJA_MD
    jinja_env = Environment(
        loader=FileSystemLoader(jinja_md_dir), extensions=[tres_jinja.MDSectionInclude, tres_jinja.MDCleanInclude]
    )
    if expect_exception:
        with pytest.raises(Exception):
            _ = jinja_env.get_template(template)
        return
    template = jinja_env.get_template(template)
    output_str = template.render()
    expected_f = jinja_md_dir / expected
    expected_content = expected_f.open('r').read()
    assert output_str == expected_content


def test_jinja_datestamp_default(testdata_dir: pathlib.Path) -> None:
    """Test jinja datestamp generation, default case."""
    jinja_md_dir = testdata_dir / JINJA_MD
    jinja_env = Environment(loader=FileSystemLoader(jinja_md_dir), extensions=[tres_jinja.MDDatestamp])
    template = jinja_env.get_template('MDDatestamp_default.jinja.md')
    output_str = template.render()
    expected_content = date.today().strftime(markdown_const.JINJA_DATESTAMP_FORMAT) + '\n\n\n# Heading'
    assert output_str == expected_content


def test_jinja_datestamp_newline(testdata_dir: pathlib.Path) -> None:
    """Test jinja datestamp generation, with newline true and false."""
    jinja_md_dir = testdata_dir / JINJA_MD
    jinja_env = Environment(loader=FileSystemLoader(jinja_md_dir), extensions=[tres_jinja.MDDatestamp])
    template = jinja_env.get_template('MDDatestamp_newline.jinja.md')
    output_str = template.render()
    expected_content = date.today().strftime(markdown_const.JINJA_DATESTAMP_FORMAT) + date.today().strftime(
        markdown_const.JINJA_DATESTAMP_FORMAT
    ) + '\n\n\n# Heading'
    assert output_str == expected_content


def test_jinja_datestamp_format(testdata_dir: pathlib.Path) -> None:
    """Test jinja datestamp generation, with format string."""
    jinja_md_dir = testdata_dir / JINJA_MD
    jinja_env = Environment(loader=FileSystemLoader(jinja_md_dir), extensions=[tres_jinja.MDDatestamp])
    template = jinja_env.get_template('MDDatestamp_format.jinja.md')
    output_str = template.render()
    expected_content = date.today().strftime('%B %d, %Y') + '\n\n\n# Heading'
    assert output_str == expected_content


def test_jinja_datestamp_invalid(testdata_dir: pathlib.Path) -> None:
    """Test jinja datestamp generation, with format string."""
    jinja_md_dir = testdata_dir / JINJA_MD
    jinja_env = Environment(loader=FileSystemLoader(jinja_md_dir), extensions=[tres_jinja.MDDatestamp])
    with pytest.raises(Exception):
        _ = jinja_env.get_template('MDDatestamp_invalid.jinja.md')



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for dealing with models."""
import logging
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import trestle.common.err as err

import typing_extensions

logger = logging.getLogger(__name__)


def get_origin(field_type: Type[Any]) -> Optional[Type[Any]]:
    """Generalized and robust get_origin function.

    This function is derived from work by pydantic, however, avoids complications
    from various python versions.
    """
    # This executes a fallback that allows a list to be generated from a constrained list.
    return typing_extensions.get_origin(field_type) or getattr(field_type, '__origin__', None)


def _get_model_field_info(field_type: Type[Any]) -> Tuple[Optional[Type[Any]], Optional[str], Optional[Type[Any]]]:
    """Need special handling for pydantic wrapped __root__ objects."""
    # oscal_read of roles.json yields pydantic.Roles model with __root__ containing list of Role
    root: Optional[Type[Any]] = None
    root_type: Optional[str] = None
    singular_type: Optional[Type[Any]] = None
    try:
        fields = field_type.__fields__
        root = fields['__root__']
        singular_type = root.type_
        root_type = root.outer_type_._name
    except Exception:  # noqa S110
        pass
    return root, root_type, singular_type


def is_collection_field_type(field_type: Type[Any]) -> bool:
    """Check whether a type hint is a collection type as used by OSCAL.

    Specifically this is whether the type is a list or not.

    Args:
        field_type: A type or a type alias of a field typically as served via pydantic introspection

    Returns:
        True if it is a collection type list.
    """
    # first check if it is a pydantic __root__ object
    _, root_type, _ = _get_model_field_info(field_type)
    if root_type == 'List':
        return True
    # Retrieves type from a type annotation
    origin_type = get_origin(field_type)
    return origin_type == list


def get_inner_type(collection_field_type: Union[Type[List[Any]], Type[Dict[str, Any]]]) -> Type[Any]:
    """Get the inner model in a generic collection model such as a List or a Dict.

    For a dict the return type is of the value and not the key.

    Args:
        collection_field_type: Provided type annotation from a pydantic object

    Returns:
        The desired type.
    """
    try:
        # Pydantic special cases must be dealt with here:
        _, _, singular_type = _get_model_field_info(collection_field_type)
        if singular_type is not None:
            return singular_type
        return typing_extensions.get_args(collection_field_type)[-1]
    except Exception as e:
        logger.debug(e)
        raise err.TrestleError('Model type is not a Dict or List') from e



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the component author module."""

import pathlib
import shutil
from typing import Any, Dict, Tuple

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils

import trestle.core.generic_oscal as generic
import trestle.oscal.component as comp
from trestle.common import const, file_utils
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.commands.href import HrefCmd
from trestle.core.control_interface import ControlInterface
from trestle.core.markdown.markdown_processor import MarkdownProcessor

md_path = 'md_comp'


def edit_files(control_path: pathlib.Path, set_parameters_flag: bool, guid_dict: Dict[str, str]) -> None:
    """Edit the files to show assemble worked."""
    assert control_path.exists()
    assert file_utils.insert_text_in_file(control_path, None, guid_dict['text'])
    if set_parameters_flag:
        assert test_utils.delete_line_in_file(control_path, 'label:')
        assert file_utils.insert_text_in_file(control_path, 'ac-1_prm_1:', '    label: label from edit\n')
        # delete profile values for 4, then replace value for 3 with new value
        assert file_utils.insert_text_in_file(control_path, 'officer', '    profile-values: new value\n')
        assert test_utils.delete_line_in_file(control_path, 'weekly')


def load_file(trestle_root: pathlib.Path, source_name: str, dest_name: str, source_type: str) -> None:
    """Load file into workspace."""
    item_orig_path = test_utils.JSON_TEST_DATA_PATH / (source_name)
    item_dir = trestle_root / f'{source_type}s/{dest_name}'
    item_dir.mkdir(exist_ok=True, parents=True)
    item_new_path = item_dir / f'{source_type}.json'
    shutil.copy(item_orig_path, item_new_path)


def setup_component_generate(trestle_root: pathlib.Path) -> Tuple[pathlib.Path, pathlib.Path, pathlib.Path]:
    """Set up files for profile generate."""
    comp_name = 'test_comp'
    cat_name = 'nist_cat'
    prof_name = 'nist_prof'
    simple_cat_name = 'simple_catalog_no_parts'
    load_file(trestle_root, 'comp_def.json', comp_name, 'component-definition')
    load_file(trestle_root, test_utils.SIMPLIFIED_NIST_CATALOG_NAME, cat_name, 'catalog')
    load_file(trestle_root, simple_cat_name + '.json', simple_cat_name, 'catalog')
    load_file(trestle_root, test_utils.SIMPLIFIED_NIST_PROFILE_NAME, prof_name, 'profile')
    new_href = 'trestle://catalogs/nist_cat/catalog.json'
    assert HrefCmd.change_import_href(trestle_root, prof_name, new_href, 0) == 0
    return comp_name, prof_name, cat_name


def check_common_contents(header: Dict[str, Any]) -> None:
    """Check common features of controls markdown."""
    params = header[const.RULES_PARAMS_TAG]
    assert len(params) == 1
    assert params[0] == {
        'name': 'foo_length', 'description': 'minimum_foo_length', 'rule-id': 'XCCDF', 'options': '["6", "9"]'
    }
    assert header[const.TRESTLE_GLOBAL_TAG][
        const.PROFILE_TITLE] == 'NIST Special Publication 800-53 Revision 5 MODERATE IMPACT BASELINE'  # noqa E501


def check_ac1_contents(ac1_path: pathlib.Path) -> None:
    """Check the contents of ac-1 md."""
    assert test_utils.confirm_text_in_file(ac1_path, 'enter one of:', 'set to 644')
    assert test_utils.confirm_text_in_file(ac1_path, 'set to 644', 'Status: implemented')
    assert test_utils.confirm_text_in_file(ac1_path, 'Status: implemented', 'ac1 remark')
    assert test_utils.confirm_text_in_file(ac1_path, 'ac-1_smt.c', 'Status: planned')
    markdown_processor = MarkdownProcessor()
    header, _ = markdown_processor.read_markdown_wo_processing(ac1_path)
    assert header[const.PARAM_VALUES_TAG]['ac-1_prm_1'] == 'Param_1_value_in_catalog'
    rules = header[const.COMP_DEF_RULES_TAG]
    assert len(rules) == 1
    assert rules[0] == {'name': 'XCCDF', 'description': 'The XCCDF must be compliant'}
    vals = header[const.COMP_DEF_RULES_PARAM_VALS_TAG]
    assert len(vals) == 2
    assert vals['quantity_available'] == '500'
    check_common_contents(header)


def check_ac5_contents(ac5_path: pathlib.Path) -> None:
    """Check the contents of ac-5 md."""
    markdown_processor = MarkdownProcessor()
    header, _ = markdown_processor.read_markdown_wo_processing(ac5_path)
    assert test_utils.confirm_text_in_file(
        ac5_path, '### Implementation Status: partial', '### Implementation Status Remarks: this is my remark'
    )
    rules = header[const.COMP_DEF_RULES_TAG]
    assert len(rules) == 2
    assert rules[0] == {'name': 'XCCDF', 'description': 'The XCCDF must be compliant'}
    assert rules[1] == {'name': 'FancyXtraRule', 'description': 'This is a fancy extra rule'}
    check_common_contents(header)


def test_component_generate(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test component generate."""
    comp_name, _, _ = setup_component_generate(tmp_trestle_dir)
    ac1_path = tmp_trestle_dir / f'{md_path}/OSCO/ac/ac-1.md'
    ac5_path = tmp_trestle_dir / f'{md_path}/OSCO/ac/ac-5.md'

    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'

    # generate the md first time
    test_utils.execute_command_and_assert(generate_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    check_ac1_contents(ac1_path)
    check_ac5_contents(ac5_path)

    file_checker = test_utils.FileChecker(tmp_trestle_dir / md_path)

    # generate again but force use of source in comp_def to load profile rather than command line
    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'
    # confirm it overwrites existing md properly
    test_utils.execute_command_and_assert(generate_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)

    # all files should be the same
    assert file_checker.files_unchanged()

    # make edits to status and remarks and control level prose
    assert test_utils.substitute_text_in_file(ac1_path, '644', '567')
    control_prose = '567 or more restrictive'
    assert test_utils.confirm_text_in_file(ac1_path, 'after assembly to JSON', control_prose)
    assert test_utils.substitute_text_in_file(ac5_path, 'Status: partial', 'Status: implemented')
    assert test_utils.confirm_text_in_file(ac5_path, 'garbage collection', 'Status: implemented')
    assert test_utils.substitute_text_in_file(ac5_path, 'my remark', 'my new remark')
    assert test_utils.confirm_text_in_file(ac5_path, 'garbage collection', 'my new remark')

    assemble_cmd = f'trestle author component-assemble -n {comp_name} -o assem_comp -m {md_path}'
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)

    # confirm assembled model is as expected
    assem_comp_def, orig_comp_def_path = ModelUtils.load_top_level_model(
        tmp_trestle_dir,
        'assem_comp',
        comp.ComponentDefinition
    )
    component = ControlInterface.get_component_by_name(assem_comp_def, 'OSCO')
    imp_reqs = ControlInterface.get_control_imp_reqs(component.control_implementations[1], 'ac-5')
    new_status = ControlInterface.get_status_from_props(imp_reqs[0])
    assert new_status.state == 'implemented'
    assert new_status.remarks.__root__ == 'this is my new remark'
    imp_reqs = ControlInterface.get_control_imp_reqs(component.control_implementations[0], 'ac-1')
    assert control_prose in imp_reqs[0].description

    orig_uuid = assem_comp_def.uuid

    orig_file_creation = orig_comp_def_path.stat().st_mtime

    # confirm repeat assemble doesn't generate new file since no changes
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    assert orig_comp_def_path.stat().st_mtime == orig_file_creation

    # force overwrite of output, regenerate, and new version
    assemble_cmd = f'trestle author component-assemble -o assem_comp -m {md_path} -r -vn 1.2.3'
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    assem_comp_def, _ = ModelUtils.load_top_level_model(tmp_trestle_dir, 'assem_comp', comp.ComponentDefinition)
    assert assem_comp_def.uuid != orig_uuid
    assert assem_comp_def.metadata.version.__root__ == '1.2.3'


def test_generic_oscal() -> None:
    """Test generic oscal conversions."""
    generic_component = generic.GenericComponent.generate()
    def_comp = generic_component.as_defined_component()
    assert def_comp.description == const.REPLACE_ME

    generic_cont_imp = generic.GenericControlImplementation.generate()
    cont_imp = generic_cont_imp.as_ssp()
    assert cont_imp.description == const.REPLACE_ME



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle command related utilities."""
import pathlib
from typing import Any, List, Optional, Type, Union

from trestle.common import const, str_utils, type_utils
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.common.str_utils import AliasMode, classname_to_alias
from trestle.core.base_model import OscalBaseModel
from trestle.core.models.elements import ElementPath
from trestle.core.models.file_content_type import FileContentType


def model_type_is_too_granular(model_type: Type[Any]) -> bool:
    """Is an model_type too fine to split."""
    if type_utils.is_collection_field_type(model_type):
        return False
    if hasattr(model_type, '__fields__') and '__root__' in model_type.__fields__:
        return True
    if model_type.__name__ in ['str', 'ConstrainedStrValue', 'int', 'float', 'datetime']:
        return True
    return False


def split_is_too_fine(split_paths: str, model_obj: OscalBaseModel) -> bool:
    """Determine if the element path list goes too fine, e.g. individual strings."""
    for split_path in split_paths.split(','):
        # find model type one level above if finishing with '.*'
        model_type = ElementPath(split_path.rstrip('.*')).get_type(type(model_obj))
        if model_type_is_too_granular(model_type):
            return True
    return False


def parse_element_args(
    model: Union[OscalBaseModel, None],
    element_args: List[str],
    relative_path: Optional[pathlib.Path] = None
) -> List[ElementPath]:
    """Parse element args into a list of ElementPath.

    The element paths are either simple links of two elements, or two elements followed by *.
    The * represents either a list of the items in that element, or a splitting of that element into its parts.
    The only parts split off are the non-trivial ones determined by the granularity check.

    contextual_mode specifies if the path is a valid project model path or not. For example,
    if we are processing a metadata.parties.*, we need to know which metadata we are processing. If we pass
    contextual_mode=true, we can infer the root model by inspecting the file directory

    If contextual_mode=False, then the path must include the full path, e.g. catalog.metadata.parties.* instead of just
    metadata.parties.*

    When the * represents splitting a model rather than a list, the model is inspected for what parts are available,
    and for each new part two element paths are created, one for the parent to the current element, and another from
    the current element to the child.

    A path may have multiple *'s, but only the final one can represent splitting a model.

    Args:
        model: The OscalBaseModel being inspected to determine available elements that can be split
        element_args: List of str representing links in the chain of element paths to be parsed
        relative_path: Optional relative path (from trestle root) used to validate element args are valid.
    Returns:
        The requested parsed list of ElementPath for use in split
    """
    # collect all paths
    element_paths: List[ElementPath] = []
    for element_arg in element_args:
        paths = parse_element_arg(model, element_arg, relative_path)
        element_paths.extend(paths)

    return element_paths


def parse_chain(
    model_obj: Union[OscalBaseModel, None],
    path_parts: List[str],
    relative_path: Optional[pathlib.Path] = None
) -> List[ElementPath]:
    """Parse the model chain starting from the beginning.

    Args:
        model_obj: Model to use for inspecting available elements, if available or none
        path_parts: list of string paths to parse including wildcards
        relative_path: Optional relative path (w.r.t trestle project root directory)

    Returns:
        List of ElementPath
    """
    element_paths: List[ElementPath] = []
    sub_model = model_obj
    have_model_to_parse = model_obj is not None

    prev_element_path = None
    latest_path = None
    parent_model = path_parts[0]
    i = 1
    while i < len(path_parts):
        p = path_parts[i]

        # if hit wildcard create element path up to this point
        if p == ElementPath.WILDCARD and len(element_paths) > 0:
            # append wildcard to the latest element path
            latest_path = element_paths.pop()
            if latest_path.get_last() == ElementPath.WILDCARD:
                raise TrestleError(f'Invalid element path with consecutive {ElementPath.WILDCARD}')

            latest_path_str = ElementPath.PATH_SEPARATOR.join([latest_path.to_string(), p])
            element_path = ElementPath(latest_path_str, latest_path.get_parent())
        else:
            # create and append element_path
            # at this point sub_model may be a list of items
            # new element path is needed only if any of the items contains the desired part
            if p != ElementPath.WILDCARD:
                new_attrib = str_utils.dash_to_underscore(p)
                if isinstance(sub_model, list):
                    for item in sub_model:
                        # go into the list and find one with requested part
                        sub_item = getattr(item, new_attrib, None)
                        if sub_item is not None:
                            sub_model = sub_item
                            break
                else:
                    sub_model = getattr(sub_model, new_attrib, None)
            if have_model_to_parse and sub_model is None:
                return element_paths
            p = ElementPath.PATH_SEPARATOR.join([parent_model, p])
            element_path = ElementPath(p, parent_path=prev_element_path)

        # If the path has wildcard and there are more parts later,
        # get the parent model for the alias path
        # If path has wildcard and it does not refer to a list, then there can be nothing after *
        if element_path.get_last() == ElementPath.WILDCARD:
            full_path_str = ElementPath.PATH_SEPARATOR.join(element_path.get_full_path_parts()[:-1])
            parent_model = ModelUtils.get_singular_alias(full_path_str, relative_path)
            # Does wildcard mean we need to inspect the sub_model to determine what can be split off from it?
            # If it has __root__ it may mean it contains a list of objects and should be split as a list
            if isinstance(sub_model, OscalBaseModel):
                root = getattr(sub_model, '__root__', None)
                if root is None or not isinstance(root, list):
                    # Cannot have parts beyond * if it isn't a list
                    if i < len(path_parts) - 1:
                        raise TrestleError(
                            f'Cannot split beyond * when the wildcard does not refer to a list.  Path: {path_parts}'
                        )
                    for key in sub_model.__fields__.keys():
                        # only create element path is item is present in the sub_model
                        if getattr(sub_model, key, None) is None:
                            continue
                        new_alias = str_utils.underscore_to_dash(key)
                        new_path = full_path_str + '.' + new_alias
                        if not split_is_too_fine(new_path, model_obj):
                            # to add parts of an element, need to add two links
                            # prev_element_path may be None, for example catalog.*
                            if prev_element_path is not None:
                                element_paths.append(prev_element_path)
                            element_paths.append(ElementPath(parent_model + '.' + new_alias, latest_path))
                    # Since wildcard is last in the chain when splitting an oscal model we are done
                    return element_paths
        else:
            parent_model = element_path.get_element_name()

        # store values for next cycle
        prev_element_path = element_path
        element_paths.append(element_path)
        i += 1
    return element_paths


def parse_element_arg(
    model_obj: Union[OscalBaseModel, None],
    element_arg: str,
    relative_path: Optional[pathlib.Path] = None
) -> List[ElementPath]:
    """Parse an element arg string into a list of ElementPath.

    Args:
        model_obj: The OscalBaseModel being inspected to determine available elements that can be split
        element_arg: Single element path, as a string.
        relative_path: Optional relative path (from trestle root) used to validate element args are valid.
    Returns:
        The requested parsed list of ElementPath for use in split
    """
    element_arg = element_arg.strip()

    if element_arg == '*':
        raise TrestleError('Invalid element path containing only a single wildcard.')

    if element_arg == '':
        raise TrestleError('Invalid element path is empty string.')

    # search for wildcards and create paths with its parent path
    path_parts = element_arg.split(ElementPath.PATH_SEPARATOR)
    if len(path_parts) <= 1:
        raise TrestleError(f'Invalid element path "{element_arg}" with only one element and no wildcard')

    element_paths = parse_chain(model_obj, path_parts, relative_path)

    if len(element_paths) <= 0:
        # don't complain if nothing to split
        pass

    return element_paths


def to_model_file_name(model_obj: OscalBaseModel, file_prefix: str, content_type: FileContentType) -> str:
    """Return the file name for the item."""
    file_ext = FileContentType.to_file_extension(content_type)
    model_type = classname_to_alias(type(model_obj).__name__, AliasMode.JSON)
    file_name = f'{file_prefix}{const.IDX_SEP}{model_type}{file_ext}'
    return file_name



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle cmd utils module."""
from typing import List

import pytest

from tests import test_utils

from trestle.common.err import TrestleError
from trestle.core.commands.common import cmd_utils
from trestle.core.models.elements import ElementPath


def prepare_expected_element_paths(element_args: List[str]) -> List[ElementPath]:
    """Prepare a list of ElementPath from a list of path strings."""
    element_paths: List[ElementPath] = []
    for element_arg in element_args:
        element_paths.append(ElementPath(element_arg))

    return element_paths


def test_parse_element_arg(tmp_path, keep_cwd):
    """Unit test parse a single element arg."""
    with pytest.raises(TrestleError):
        cmd_utils.parse_element_arg(None, 'component-definition')

    with pytest.raises(TrestleError):
        cmd_utils.parse_element_arg(None, '*.*')

    with pytest.raises(TrestleError):
        cmd_utils.parse_element_arg(None, '*')

    with pytest.raises(TrestleError):
        cmd_utils.parse_element_arg(None, 'component-definition.components.*.*')

    element_arg = 'catalog.groups.*'
    expected_paths: List[ElementPath] = prepare_expected_element_paths(['catalog.groups.*'])
    element_paths: List[ElementPath] = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    # contextual path
    test_utils.ensure_trestle_config_dir(tmp_path)
    catalog_split_dir = tmp_path / 'catalogs/nist800-53/catalog'
    catalog_split_dir.mkdir(exist_ok=True, parents=True)
    rel_dir = catalog_split_dir.relative_to(tmp_path)
    element_arg = 'groups.*'
    expected_paths = prepare_expected_element_paths(['groups.*'])
    element_paths = cmd_utils.parse_element_arg(None, element_arg, rel_dir)
    assert expected_paths == element_paths

    element_arg = 'catalog.metadata.parties.*'
    expected_paths = prepare_expected_element_paths(['catalog.metadata', 'metadata.parties.*'])
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'component-definition.components'
    expected_paths = prepare_expected_element_paths(['component-definition.components'])
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'component-definition.components.*'
    expected_paths = prepare_expected_element_paths(['component-definition.components.*'])
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'component-definition.metadata.roles'
    expected_paths = prepare_expected_element_paths(['component-definition.metadata', 'metadata.roles'])
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'catalog.groups.*.controls.*.controls.*'
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('group.controls.*', parent_path=p1)
    p3 = ElementPath('control.controls.*', parent_path=p2)
    expected_paths = [p1, p2, p3]
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'catalog.groups.*.controls'
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('group.controls', parent_path=p1)
    expected_paths = [p1, p2]
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'component-definition.components.*.control-implementations'
    p1 = ElementPath('component-definition.components.*')
    p2 = ElementPath('defined-component.control-implementations', parent_path=p1)
    expected_paths = [p1, p2]
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    element_arg = 'component-definition.components.*.control-implementations.*'
    p1 = ElementPath('component-definition.components.*')
    p2 = ElementPath('defined-component.control-implementations.*', parent_path=p1)
    expected_paths = [p1, p2]
    element_paths = cmd_utils.parse_element_arg(None, element_arg)
    assert expected_paths == element_paths

    # use contextual path for parsing path
    test_utils.ensure_trestle_config_dir(tmp_path)
    comp_def_dir = tmp_path / 'component-definitions/mycomponent/'
    comp_def_dir.mkdir(exist_ok=True, parents=True)
    rel_dir = comp_def_dir.relative_to(tmp_path)
    element_arg = 'metadata.parties.*'
    expected_paths: List[ElementPath] = prepare_expected_element_paths(['metadata.parties.*'])
    element_paths: List[ElementPath] = cmd_utils.parse_element_arg(None, element_arg, rel_dir)
    assert expected_paths == element_paths


def test_parse_element_args():
    """Unit test parse multiple element args."""
    element_args = ['catalog.metadata', 'catalog.groups', 'catalog.controls']
    p0 = ElementPath('catalog.metadata')
    p1 = ElementPath('catalog.groups')
    p2 = ElementPath('catalog.controls')
    expected_paths: List[ElementPath] = [p0, p1, p2]
    element_paths = cmd_utils.parse_element_args(None, element_args, False)
    assert expected_paths == element_paths

    # element args with wildcard
    element_args = ['catalog.metadata', 'catalog.groups.*.controls.*.controls.*', 'catalog.controls.*.controls.*']
    p0 = ElementPath('catalog.metadata')
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('group.controls.*', parent_path=p1)
    p3 = ElementPath('control.controls.*', parent_path=p2)
    p4 = ElementPath('catalog.controls.*')
    p5 = ElementPath('control.controls.*', parent_path=p4)
    expected: List[ElementPath] = [p0, p1, p2, p3, p4, p5]
    assert cmd_utils.parse_element_args(None, element_args, False) == expected


@pytest.mark.parametrize('element_arg', ['catalog.metadata.*', 'catalog.*'])
def test_parse_element_args_split_model(element_arg, simplified_nist_catalog):
    """Test split of model with wildcard."""
    element_paths = cmd_utils.parse_element_arg(simplified_nist_catalog, element_arg, False)
    assert element_paths


@pytest.mark.parametrize(
    'element_arg',
    [
        ('component-definition.foo', []),
        ('component-definition.*', [['component-definition', 'components'], ['component-definition', 'metadata']]),
        ('component-definition.components', [['component-definition', 'components']]),
        ('component-definition.components.*', [['component-definition', 'components', '*']]),
        (
            'component-definition.components.*.roles',
            [['component-definition', 'components', '*'], ['defined-component', 'roles']]
        )
    ]
)
def test_parse_element_args_split_compdef(element_arg, sample_component_definition):
    """Test split of model with wildcard."""
    element_paths = cmd_utils.parse_element_arg(sample_component_definition, element_arg[0], False)
    assert len(element_paths) == len(element_arg[1])
    for ii in range(len(element_paths)):
        assert element_paths[ii]._path in element_arg[1]


@pytest.mark.parametrize('element_arg', ['catalog.metadata.*.roles', 'catalog', '', '*'])
def test_parse_element_arg_split_model_failures(element_arg, simplified_nist_catalog):
    """Test failures split of model with wildcard."""
    with pytest.raises(TrestleError):
        cmd_utils.parse_element_arg(simplified_nist_catalog, element_arg, False)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle remove path action class."""

import pathlib

import pytest

from tests import test_utils

import trestle.common.const as const
from trestle.common import trash
from trestle.common.err import TrestleError
from trestle.core.models.actions import RemovePathAction


def test_remove_path_file(tmp_path: pathlib.Path) -> None:
    """Test remove path with content clear option."""
    tmp_data_dir = tmp_path.joinpath('data')
    tmp_data_file = tmp_data_dir.joinpath('readme.md')
    tmp_data_dir.mkdir(exist_ok=True, parents=True)

    # not a valid trestle project should error in constructor
    with pytest.raises(TrestleError):
        rpa = RemovePathAction(tmp_data_file)

    # create trestle project
    test_utils.ensure_trestle_config_dir(tmp_path)
    rpa = RemovePathAction(tmp_data_file)

    # missing file should error
    # with pytest.raises(TrestleError): # noqa: E800  remove path is not working properly
    #    rpa.execute()                  # noqa: E800

    # write some content in the file
    file_pos = 0
    dummy_data = 'DUMMY DATA'
    with open(tmp_data_file, 'a+', encoding=const.FILE_ENCODING) as fp:
        fp.write(dummy_data)
        file_pos = fp.tell()
    assert file_pos >= 0

    # remove file
    tmp_data_file_trash = trash.to_trash_file_path(tmp_data_file)
    assert tmp_data_file_trash.exists() is False
    rpa.execute()
    tmp_data_file_trash.exists()
    assert tmp_data_file.exists() is False

    # rollback file
    rpa.rollback()
    tmp_data_file_trash.exists()
    tmp_data_file.exists()
    with open(tmp_data_file, 'a+', encoding=const.FILE_ENCODING) as fp:
        assert file_pos == fp.tell()

    # remove dir
    rpa = RemovePathAction(tmp_data_dir)
    tmp_data_trash = trash.to_trash_dir_path(tmp_data_dir)
    tmp_data_file_trash = trash.to_trash_file_path(tmp_data_file)
    if tmp_data_trash.exists():
        tmp_data_trash.rmdir()
    rpa.execute()
    assert tmp_data_trash.exists()
    assert tmp_data_file_trash.exists()
    assert tmp_data_file.exists() is False
    assert tmp_data_dir.exists() is False

    # rollback dir
    rpa.rollback()
    assert tmp_data_trash.exists() is False
    assert tmp_data_file_trash.exists() is False
    assert tmp_data_file.exists()
    assert tmp_data_dir.exists()


def test_remove_path_magic_methods(tmp_path):
    """Test remove path magic methods."""
    tmp_data_dir = tmp_path.joinpath('data')
    test_utils.ensure_trestle_config_dir(tmp_path)
    rpa = RemovePathAction(tmp_data_dir)

    action_desc = rpa.to_string()
    assert action_desc == f'{rpa.get_type()} {tmp_data_dir}'
